import gc
import logging
import os
import shutil
import smtplib
import time
import socket
from argparse import ArgumentParser
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path

import requests
from apscheduler.executors.pool import ThreadPoolExecutor, ProcessPoolExecutor

from sherlockpipe.bayesian_fit.run import run_fit
from sherlockpipe.search.run import run_search, load_from_yaml
from sherlockpipe.stability import stability_args_parse
from sherlockpipe.system_stability.run import run_stability
from sherlockpipe.validate import validation_args_parse
from sherlockpipe.validation.run import run_validate
from sherlockpipe.vetting.run import run_vet


def move_input_to_working(running_file, working_dir):
    filename = os.path.basename(running_file)
    working_file = working_dir + '/' + filename
    shutil.move(running_file, working_file)
    return working_file

def run_script(input_dir, output_dir, working_dir, cpus, pa):
    running_file = None
    working_file = None
    working_target_dir = None
    properties = None
    try:
        paths = sorted(Path(input_dir).iterdir(), key=os.path.getmtime)
        for index, path in enumerate(paths):
            path_str = str(path)
            if path_str.endswith(".yaml") or path_str.endswith('.yml'):
                running_file = path_str
                working_file = move_input_to_working(running_file, working_dir)
                working_target_dir = working_dir + '/' + Path(working_file).stem
                os.mkdir(working_target_dir)
                os.chdir(working_target_dir)
                properties = run_search(working_file, False, None, cpus=cpus)
            elif os.path.isdir(path_str):
                running_file = path_str
                working_target_dir = move_input_to_working(path_str, working_dir)
                vet_file = working_target_dir + '/vet.yaml'
                validate_file = working_target_dir + '/validate.yaml'
                fit_file = working_target_dir + '/fit.yaml'
                plan_file = working_target_dir + '/plan.yaml'
                stability_file = working_target_dir + '/stability.yaml'
                if os.path.exists(vet_file):
                    properties = load_from_yaml(vet_file)
                    candidate = None if 'CANDIDATE' not in properties else properties['CANDIDATE']
                    run_vet(working_file, candidate, properties, cpus)
                if os.path.exists(validate_file):
                    properties = load_from_yaml(validate_file)
                    candidate = None if 'CANDIDATE' not in properties else properties['CANDIDATE']
                    contrast_curve = None if 'CONTRAST_CURVE' not in properties else properties['CONTRAST_CURVE']
                    bins = None if 'BINS' not in properties else properties['BINS']
                    scenarios = None if 'SCENARIOS' not in properties else properties['SCENARIOS']
                    sigma_mode = None if 'SIGMA_MODE' not in properties else properties['SIGMA_MODE']
                    args = validation_args_parse({'object_dir': path_str, 'candidate': candidate,
                                                  'properties': properties, 'contrast_curve': contrast_curve,
                                                  'bins': bins, 'scenarios': scenarios, 'sigma_mode': sigma_mode})
                    run_validate(args)
                if os.path.exists(fit_file):
                    properties = load_from_yaml(fit_file)
                    candidate = None if 'CANDIDATE' not in properties else properties['CANDIDATE']
                    only_initial = None if 'ONLY_INITIAL' not in properties else properties['ONLY_INITIAL']
                    tolerance = None if 'TOLERANCE' not in properties else properties['TOLERANCE']
                    mcmc = None if 'MCMC' not in properties else properties['MCMC']
                    detrend = None if 'DETREND' not in properties else properties['DETREND']
                    fit_orbit = None if 'FIT_ORBIT' not in properties else properties['FIT_ORBIT']
                    args = validation_args_parse({'object_dir': path_str, 'candidate': candidate,
                                                  'properties': properties, 'only_initial': only_initial,
                                                  'tolerance': tolerance, 'mcmc': mcmc, 'detrend': detrend,
                                                  'fit_orbit': fit_orbit})
                    run_fit(args)
                if os.path.exists(plan_file):
                    properties = load_from_yaml(vet_file)
                if os.path.exists(stability_file):
                    properties = load_from_yaml(vet_file)
                    max_ecc = None if 'MAX_ECC' not in properties else properties['MAX_ECC']
                    period_bins = None if 'PERIOD_BINS' not in properties else properties['PERIOD_BINS']
                    ecc_bins = None if 'ECC_BINS' not in properties else properties['ECC_BINS']
                    inc_bins = None if 'INC_BINS' not in properties else properties['INC_BINS']
                    omega_bins = None if 'OMEGA_BINS' not in properties else properties['OMEGA_BINS']
                    mass_bins = None if 'MASS_BINS' not in properties else properties['MASS_BINS']
                    star_mass_bins = None if 'STAR_MASS_BINS' not in properties else properties['STAR_MASS_BINS']
                    years = None if 'YEARS' not in properties else properties['YEARS']
                    spock = None if 'SPOCK' not in properties else properties['SPOCK']
                    free_params = None if 'CANDIDATE' not in properties else properties['CANDIDATE']
                    args = stability_args_parse({'object_dir': path_str, 'max_ecc': max_ecc,
                                                  'period_bins': period_bins, 'ecc_bins': ecc_bins,
                                                  'inc_bins': inc_bins, 'omega_bins': omega_bins,
                                                  'mass_bins': mass_bins, 'star_mass_bins': star_mass_bins,
                                                  'years': years, 'spock': spock,
                                                  'free_params': free_params})
                    run_stability(args)
    finally:
        os.chdir(working_dir)
        if running_file is not None and os.path.exists(running_file):
            shutil.rmtree(running_file, ignore_errors=True)
        if working_file is not None and os.path.exists(working_file):
            shutil.rmtree(working_file, ignore_errors=True)
        if working_target_dir is not None and os.path.exists(working_target_dir):
            logging.info("Finished file %s", running_file)
            output_file = output_dir + '/' + Path(working_target_dir).stem
            shutil.move(working_target_dir, output_file)
            if properties is not None and 'EMAIL' in properties:
                receiver_address = properties['EMAIL']
                send_email(output_file, receiver_address, pa)
            else:
                logging.warning('Cannot send an email because the receiver email was not provided.')
        gc.collect()


def send_email(filename, receiver_address, pa):
    try:
        sender_address = 'sherlockpipeline@gmail.com'
        sender_pass = pa
        # Setup the MIME
        message = MIMEMultipart()
        message['From'] = sender_address
        message['To'] = receiver_address
        host = socket.gethostname()
        try:
            host = host if host != 'localhost' else requests.get('https://api.ipify.org').content.decode('utf8')
        except:
            print('Cant identify host in the internet')
        message['Subject'] = 'SHERLOCK [' + host + ']: Job with filename ' + os.path.basename(filename) + ' finished'
        # The body and the attachments for the mail
        message.attach(MIMEText('You can check all the target directories under the directory: ' + filename +
                                '. Keep in mind that these files will be kept only for two weeks and then will be '
                                ' automatically removed.', 'plain'))
        # Create SMTP session for sending the mail
        session = smtplib.SMTP('smtp.gmail.com', 587)  # use gmail with port
        session.starttls()  # enable security
        session.login(sender_address, sender_pass)  # login with mail_id and password
        text = message.as_string()
        session.sendmail(sender_address, receiver_address, text)
        session.quit()
        logging.info('Sent email to %s for job %s', receiver_address, filename)
    except Exception:
        logging.exception('Cant send email to %s for job %s', receiver_address, filename)


def clean_old_outputs(output_dir):
    current_time = time.time()
    max_days = 14
    files = os.listdir(output_dir)
    for file in files:
        file_path = os.path.join(output_dir, file)
        file_mod_time = os.path.getmtime(file_path)
        time_diff = current_time - file_mod_time
        if time_diff > max_days * 24 * 60 * 60:
            shutil.rmtree(file_path, ignore_errors=True)


if __name__ == '__main__':
    ap = ArgumentParser(description='Sherlock As A Service')
    ap.add_argument('--input_dir',
                    help="Drop-in directory for users",
                    required=True, type=str)
    ap.add_argument('--working_dir',
                    help="Directory to be used when working with a file",
                    required=True, type=str)
    ap.add_argument('--output_dir',
                    help="Directory where SHERLOCK will drop the results",
                    required=True, type=str)
    ap.add_argument('--cpus',
                    help="Number of cpu processes to be used",
                    required=True, type=int)
    ap.add_argument('--pa',
                    help="",
                    required=True, type=str)
    args = ap.parse_args()
    paths = sorted(Path(args.working_dir).iterdir(), key=os.path.getmtime)
    for index, path in enumerate(paths):
        file_in_working = str(path)
        if file_in_working.endswith(".yaml") or file_in_working.endswith('.yml'):
            filename = os.path.basename(file_in_working)
            shutil.move(file_in_working, args.input_dir + '/' + filename)
        else:
            shutil.rmtree(file_in_working, ignore_errors=True)
    if not isinstance(logging.root, logging.RootLogger):
        logging.root = logging.RootLogger(logging.INFO)
    from apscheduler.schedulers.blocking import BlockingScheduler
    logging.getLogger('apscheduler.executors.default').setLevel(logging.ERROR)
    executors = {
        'default': ProcessPoolExecutor(1),
        'processpool': ProcessPoolExecutor(1)
    }
    scheduler = BlockingScheduler(executors)
    scheduler.add_job(run_script, 'interval', minutes=1, max_instances=1, misfire_grace_time=600000,
                      kwargs={"input_dir": args.input_dir, "output_dir": args.output_dir,
                              "working_dir": args.working_dir, "cpus": args.cpus, "pa": args.pa}, replace_existing=True)
    scheduler.add_job(clean_old_outputs, 'interval', days=1, max_instances=1,
                      kwargs={"output_dir": args.output_dir}, replace_existing=True)
    scheduler.start()



import json
import logging
import multiprocessing
from typing import List, Dict

import matplotlib.pyplot as plt
from abc import ABC, abstractmethod
from multiprocessing import Pool
import numpy as np
import rebound
from astropy import units as u
from lcbuilder.helper import LcbuilderHelper

from sherlockpipe import constants


"""Includes classes to be used ase base for stability simulations"""


class PlanetInput:
    """
    Defines the planet parameters for system stability calculations
    """
    def __init__(self, period, period_low_err, period_up_err,
                 radius, radius_low_err, radius_up_err, eccentricity, ecc_low_err, ecc_up_err,
                 inclination, inc_low_err, inc_up_err, omega, omega_low_err, omega_up_err,
                 mass=None, mass_low_err=None, mass_up_err=None, period_bins=3,
                 mass_bins=3, ecc_bins=3, inc_bins=3, omega_bins=3):
        self.period = period
        self.period_low_err = period_low_err
        self.period_up_err = period_up_err
        self.radius = radius
        self.radius_low_err = radius_low_err
        self.radius_up_err = radius_up_err
        self.eccentricity = eccentricity
        self.ecc_low_err = ecc_low_err
        self.ecc_up_err = ecc_up_err
        self.inclination = inclination
        self.inc_low_err = inc_low_err
        self.inc_up_err = inc_up_err
        self.omega = omega
        self.omega_low_err = omega_low_err
        self.omega_up_err = omega_up_err
        self.mass = mass
        self.mass_low_err = mass_low_err
        self.mass_up_err = mass_up_err
        self.mass_bins = mass_bins if mass_bins is not None else 3
        self.period_bins = period_bins if period_bins is not None else 3
        self.ecc_bins = ecc_bins if ecc_bins is not None else 3
        self.inc_bins = inc_bins if inc_bins is not None else 3
        self.omega_bins = omega_bins if omega_bins is not None else 3


class SimulationInput:
    """
    Used as input for the simulations done for each scenario
    """

    def __init__(self, star_mass, mass_arr, planet_periods, ecc_arr, inc_arr, omega_arr, index):
        self.star_mass = star_mass
        self.mass_arr = mass_arr
        self.planet_periods = planet_periods
        self.ecc_arr = ecc_arr
        self.inc_arr = inc_arr
        self.omega_arr = omega_arr
        self.index = index


class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types. Got from https://stackoverflow.com/a/49677241/4198726"""

    def default(self, obj):
        """
        Default method invoked to convert the object

        :param obj: the input object
        :return: the json encoded object
        """
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class StabilityCalculator(ABC):
    """
    Template class for system stability calculation algorithms
    """

    def __init__(self):
        pass

    @staticmethod
    def mass_from_radius(radius):
        """
        Computation of mass-radius relationship from
        Bashi D., Helled R., Zucker S., Mordasini C., 2017, A&A, 604, A83. doi:10.1051/0004-6361/201629922

        :param radius: the radius value in earth radius
        :return: the mass in earth masses
        """
        return radius ** (1 / 0.55) if radius <= 12.1 else radius ** (1 / 0.01)

    @staticmethod
    def prepare_star_masses(star_mass_low, star_mass_up, star_mass_bins):
        """
        Creates a star masses grid

        :param star_mass_low: the lowest star mass value
        :param star_mass_up: the highest star mass value
        :param star_mass_bins: the number of star masses to sample. It will be ignored if star_mass_low == star_mass_up.
        :return: the star masses grid
        """
        return np.linspace(star_mass_low, star_mass_up, star_mass_bins) if star_mass_low != star_mass_up \
            else np.linspace(star_mass_low, star_mass_up, 1)

    @staticmethod
    def prepare_planet_params(planet_params: List[PlanetInput]):
        """
        Fills the planet masses if missing

        :param planet_params: the planet inputs
        :return: the planet inputs with the filled masses
        """
        for planet_param in planet_params:
            if planet_param.radius is None and (planet_param.mass_low_err is None or planet_param.mass_low_err is None):
                raise ValueError("There is one body without either radius or mass information: " +
                                 json.dumps(planet_param.__dict__))
            if planet_param.radius is not None:
                planet_param.mass = StabilityCalculator.mass_from_radius(planet_param.radius)
                planet_param.mass_low_err = (planet_param.mass - StabilityCalculator.mass_from_radius(planet_param.radius - planet_param.radius_low_err)) * 2
                planet_param.mass_up_err = (StabilityCalculator.mass_from_radius(planet_param.radius + planet_param.radius_up_err) - planet_param.mass) * 2
        return planet_params

    @staticmethod
    def init_rebound_simulation(simulation_input):
        """
        Initializes the simulation for rebound-based algorithms

        :param simulation_input: the input data for the simulation scenario
        :return: the rebound initialized simulation scenario
        """
        sim = rebound.Simulation()
        sim.units = ('yr', 'AU', 'Msun')
        sim.integrator = "whfast"
        sim.ri_whfast.safe_mode = 0
        sim.add(m=simulation_input.star_mass)
        min_period = 100
        for planet_key, mass in enumerate(simulation_input.mass_arr):
            period = simulation_input.planet_periods[planet_key]
            min_period = min_period if period > min_period else period
            ecc = simulation_input.ecc_arr[planet_key]
            inc = np.deg2rad(simulation_input.inc_arr[planet_key])
            omega = np.deg2rad(simulation_input.omega_arr[planet_key])
            sim.add(m=LcbuilderHelper.convert_from_to(mass, u.M_earth, u.M_sun) / simulation_input.star_mass,
                    P=period, e=ecc, omega=omega, inc=inc)
        sim.dt = min_period / 365.25 / 100
        # sim.status()
        sim.move_to_com()
        return sim

    def run(self, results_dir, star_mass_low, star_mass_up, star_mass_bins, planet_params: List[PlanetInput],
            cpus=multiprocessing.cpu_count() - 1, free_params=None):
        """
        Creates possible scenarios of stellar masses, planet masses and planet eccentricities. Afterwards a stability
        analysis is run for each of the scenarios and the results are stored in a file.

        :param results_dir: the directory where the results will be written into
        :param star_mass_low: the lowest star mass
        :param star_mass_up: the highest star mass
        :param star_mass_bins: the number of star masses to sample
        :param List[PlanetInput] planet_params: the planet inputs containing the planets parameters
        :param cpus: the number of cpus to be used
        :param free_params: the parameters to be sampled entirely
        """
        if free_params is None:
            free_params = []
        planet_params = StabilityCalculator.prepare_planet_params(planet_params)
        star_masses = StabilityCalculator.prepare_star_masses(star_mass_low, star_mass_up, star_mass_bins)
        planet_masses = []
        planet_period = []
        planet_ecc = []
        planet_inc = []
        planet_omega = []
        for planet_param in planet_params:
            if planet_param.period_bins == 1 or planet_param.period_low_err == planet_param.period_up_err == 0:
                period_grid = np.full(1, planet_param.period)
            else:
                period_grid = np.linspace(planet_param.period - planet_param.period_low_err,
                                        planet_param.period + planet_param.period_up_err,
                                        planet_param.period_bins)
            planet_period.append(period_grid)
            if planet_param.mass_bins == 1 or planet_param.mass_low_err == planet_param.mass_up_err == 0:
                mass_grid = np.full(1, planet_param.mass)
            else:
                mass_grid = np.linspace(planet_param.mass - planet_param.mass_low_err,
                                        planet_param.mass + planet_param.mass_up_err,
                                        planet_param.mass_bins)
            planet_masses.append(mass_grid)
            if "eccentricity" in free_params:
                ecc_grid = np.linspace(0, 0.5, planet_param.ecc_bins)
            elif planet_param.ecc_bins == 1 or planet_param.ecc_low_err == planet_param.ecc_up_err == 0:
                ecc_grid = np.full(1, planet_param.eccentricity)
            else:
                low_ecc = planet_param.eccentricity - planet_param.ecc_low_err
                low_ecc = low_ecc if low_ecc > 0 else 0
                up_ecc = planet_param.eccentricity + planet_param.ecc_up_err
                up_ecc = up_ecc if up_ecc < 1 else 1
                ecc_grid = np.linspace(low_ecc, up_ecc, planet_param.ecc_bins)
            planet_ecc.append(ecc_grid)
            if planet_param.inc_bins == 1 or planet_param.inc_low_err == planet_param.inc_up_err == 0:
                inc_grid = np.full(1, planet_param.inclination)
            else:
                inc_grid = np.linspace(planet_param.inclination - planet_param.inc_low_err,
                                       planet_param.inclination + planet_param.inc_up_err,
                                       planet_param.inc_bins)
            planet_inc.append(inc_grid)
            if "omega" in free_params:
                # using arange instead of linspace because 0 and 360 are the same, so we exclude 360
                omega_grid = np.arange(0, 360, 360 // planet_param.omega_bins)
            elif planet_param.omega_bins == 1 or planet_param.omega_low_err == planet_param.omega_up_err == 0:
                omega_grid = np.full(1, planet_param.omega)
            else:
                omega_grid = np.linspace(planet_param.omega - planet_param.omega_low_err,
                                         planet_param.omega + planet_param.omega_up_err,
                                         planet_param.omega_bins)
            planet_omega.append(omega_grid)
        period_grid = np.array(np.meshgrid(*np.array(planet_period))).T.reshape(-1, len(planet_period))
        masses_grid = np.array(np.meshgrid(*np.array(planet_masses))).T.reshape(-1, len(planet_masses))
        ecc_grid = np.array(np.meshgrid(*np.array(planet_ecc))).T.reshape(-1, len(planet_ecc))
        inc_grid = np.array(np.meshgrid(*np.array(planet_inc))).T.reshape(-1, len(planet_inc))
        omega_grid = np.array(np.meshgrid(*np.array(planet_omega))).T.reshape(-1, len(planet_omega))
        simulation_inputs = []
        i = 0
        star_masses_scenario_num = len(star_masses)
        masses_scenario_num = len(masses_grid)
        period_scenario_num = len(period_grid)
        inc_scenario_num = len(inc_grid)
        ecc_scenario_num = len(ecc_grid)
        omega_scenario_num = len(omega_grid)
        scenarios_num = star_masses_scenario_num * masses_scenario_num * period_scenario_num * inc_scenario_num * \
                        ecc_scenario_num * omega_scenario_num
        logging.info("Preparing system values for all scenarios")
        logging.info("%.0f star mass scenarios.", len(star_masses))
        logging.info("%.0f bodies mass scenarios.", len(masses_grid))
        logging.info("%.0f period scenarios.", len(period_grid))
        logging.info("%.0f inclination scenarios.", len(inc_grid))
        logging.info("%.0f eccentricity scenarios.", len(ecc_grid))
        logging.info("%.0f arg of periastron scenarios.", len(omega_grid))
        logging.info("%.0f x %.0f x %.0f x %.0f x %.0f x %.0f = %.0f total scenarios.", star_masses_scenario_num,
                     masses_scenario_num, period_scenario_num, inc_scenario_num, ecc_scenario_num, omega_scenario_num,
                     scenarios_num)
        for star_mass in star_masses:
            for period_key, period_arr in enumerate(period_grid):
                for mass_key, mass_arr in enumerate(masses_grid):
                    for inc_key, inc_arr in enumerate(inc_grid):
                        for ecc_key, ecc_arr in enumerate(ecc_grid):
                            for omega_key, omega_arr in enumerate(omega_grid):
                                simulation_inputs.append(SimulationInput(star_mass, mass_arr, period_arr, ecc_arr, inc_arr, omega_arr, i))
                                i = i + 1
        logging.info("Finished preparing scenarios")
        with Pool(processes=cpus) as pool:
            simulation_results = pool.map(self.log_and_run_simulation, simulation_inputs)
        self.store_simulation_results(simulation_results, results_dir)

    def log_and_run_simulation(self, simulation_input: SimulationInput) -> dict:
        """
        Logs the simulation input and launches the simulation for the given input.

        :param SimulationInput simulation_input:
        :return dict: the resulting dictionary from the simulation
        """
        logging.info("Running scenario number %.0f: %s", simulation_input.index, json.dumps(simulation_input.__dict__,
                                                                                            cls=NumpyEncoder))
        return self.run_simulation(simulation_input)

    @abstractmethod
    def run_simulation(self, simulation_input: SimulationInput) -> dict:
        """
        Runs one stability scenario

        :param SimulationInput simulation_input: the simulation scenario parameters
        :return dict: the result of the scenario in a dictionary
        """
        pass

    @abstractmethod
    def store_simulation_results(self, simulation_results: List[Dict], results_dir: str):
        """
        Writes into disk all the final simulation results

        :param List[Dict] simulation_results: the results of the simulation for all the scenarios
        :param str results_dir: the output directory where results will be written into
        """
        pass

    def plot_simulation(self, simulation_input_df, save_dir, scenario_name, xlim=None, ylim=None):
        eccs = simulation_input_df["eccentricities"].split(",")
        eccs = [float(i) for i in eccs]
        incs = simulation_input_df["inclinations"].split(",")
        incs = [float(i) - 90 for i in incs]
        omegas = simulation_input_df["arg_periastron"].split(",")
        omegas = [float(i) for i in omegas]
        periods = simulation_input_df["periods"].split(",")
        periods = [float(i) for i in periods]
        masses = simulation_input_df["masses"].split(",")
        masses = [float(i) for i in masses]
        star_mass = simulation_input_df["star_mass"]
        sim = StabilityCalculator.init_rebound_simulation(
            SimulationInput(star_mass, masses, periods, eccs, incs, omegas, 1))
        filenames = []
        for i in range(1):
            sim.integrate(sim.t + i / 100 * 2 * np.pi)
            fig, ax = rebound.OrbitPlot(sim, color=True, unitlabel="[AU]", xlim=xlim, ylim=ylim)
            filename = save_dir + "scenario_" + scenario_name + "_" + str(i) + ".png"
            plt.savefig(filename)
            filenames.append(filename)
            plt.close(fig)
        # with imageio.get_writer(save_dir + scenario_name + '.gif', mode='I') as writer:
        #     for filename in filenames:
        #         image = imageio.imread(filename)
        #         writer.append_data(image)
        # # Remove files
        # for filename in set(filenames):
        #     os.remove(filename)



import pickle
import sys
import yaml
from pathlib import Path
import importlib.util


"""Includes common functions for the main entrypoints"""


def load_module(module_path: str) -> object:
    """
    Allows the dynamic load of a python module into the system

    :param str module_path: the module directory
    :return object: the module object
    """
    spec = importlib.util.spec_from_file_location("customs", module_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


def extract_custom_class(module_path: str) -> object:
    """
    Given a module path, it loads it dynamically and extracts its inner class, assuming that the module and the class
    have the same name.

    :param str module_path: the module directory
    :return object: the class module object
    """
    class_module = None
    if module_path is not None:
        class_module = load_module(module_path)
        class_name = Path(module_path.replace(".py", "")).name
        class_module = getattr(class_module, class_name)
        globals()[class_name] = class_module
        pickle.dumps(class_module)
        class_module = class_module()
    return class_module


def get_from_dict(target: dict, key: str) -> object:
    """
    Given a target dictionary, it retrieves the value associated to a `key`

    :param dict target: the dictionary
    :param str key: the key
    :return object: the value of the key
    """
    value = None
    if isinstance(target, dict) and key in target:
        value = target[key]
    return value


def get_from_dict_or_default(target: dict, key: str, default) -> object:
    """
    This method does the same :func:`get_from_user`, but returns a default value if the key had none.

    :param dict target: the dictionary
    :param key key: the key
    :param default: the default value if nothing is found
    :return object: the final value for the key
    """
    value = get_from_dict(target, key)
    return value if value is not None else default


def get_from_user_or_config(target: dict, user_properties: dict, key: str) -> object:
    """
    Given the root user_properties dictionary and the specific target dictionary, it loads the property with name `key`
    iteratively: first tries to load the value from `user_properties`, then it tries to load it from `target`. If
    `target` contained a value, the one from `user_properties` is overwritten.

    :param dict target: the target inner properties
    :param dict user_properties: the global properties
    :param str key: the key of the properties to be extracted
    :return object: the final value
    """
    value = get_from_dict(user_properties, key)
    value = get_from_dict_or_default(target, key, value)
    return value


def get_from_user_or_config_or_default(target: dict, user_properties: dict, key: str, default) -> object:
    """
    This method does the same as :func:`get_from_user_or_config` but also returns a default value if no value was
    found associated to the given key

    :param dict target: the target inner properties
    :param dict user_properties: the global properties
    :param str key: the key of the properties to be extracted
    :param default: the default value if none was found
    :return object: the final value
    """
    value = get_from_user_or_config(target, user_properties, key)
    return value if value is not None else default


def load_from_yaml(file: str) -> dict:
    """
    Loads a yaml file into a dictionary

    :param file: the input file name
    :return dict: the dictionary
    """
    return yaml.load(open(file), yaml.SafeLoader)



import logging
import os
import shutil
from pathlib import Path

import pandas as pd
import sys

from sherlockpipe.loading import common
from sherlockpipe.validation.validator import Validator


def run_validate(args):
    object_dir = os.getcwd() if args.object_dir is None else args.object_dir
    candidates = pd.read_csv(object_dir + "/candidates.csv")
    if args.candidate is not None:
        validation_dir = object_dir + "/validate_" + str(args.candidate)
    else:
        validation_dir = object_dir + "/validate_" + str(Path(args.properties).stem)
    if os.path.exists(validation_dir) or os.path.isdir(validation_dir):
        shutil.rmtree(validation_dir, ignore_errors=True)
    os.mkdir(validation_dir)
    file_dir = validation_dir + "/validation.log"
    if os.path.exists(file_dir):
        os.remove(file_dir)
    formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    if not isinstance(logging.root, logging.RootLogger):
        logging.root = logging.RootLogger(logging.INFO)
    logger = logging.getLogger()
    while len(logger.handlers) > 0:
        logger.handlers.pop()
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    handler = logging.FileHandler(file_dir)
    handler.setLevel(logging.INFO)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logging.info("Starting validation")
    star_df = pd.read_csv(object_dir + "/params_star.csv")
    if args.candidate is None:
        logging.info("Reading validation input from properties file: %s", args.properties)
        user_properties = common.load_from_yaml(args.properties)
        candidate = pd.DataFrame(columns=['id', 'period', 'depth', 't0', 'sectors', 'ffi', 'number', 'lc'])
        candidate = candidate.append(user_properties, ignore_index=True)
        candidate['id'] = star_df.iloc[0]["obj_id"]
    else:
        candidate_selection = int(args.candidate)
        if candidate_selection < 1 or candidate_selection > len(candidates.index):
            raise SystemExit("User selected a wrong candidate number.")
        candidates = candidates.rename(columns={'Object Id': 'id'})
        candidates['number'] = 1
        candidate = candidates.iloc[[candidate_selection - 1]]
        candidate.iloc[:]['number'] = candidate_selection
        logging.info("Selected signal number " + str(candidate_selection))
    if args.sectors is not None:
        candidate['sectors'] = args.sectors
    validator = Validator(object_dir, validation_dir, len(candidate) == 1, candidates)
    validator.validate(candidate, star_df.iloc[0], args.cpus, args.contrast_curve, args.bins, args.scenarios,
                       args.sigma_mode)



import dataclasses
import logging
from multiprocessing import Pool
from typing import List

import numpy as np
import matplotlib.pyplot as plt
from scipy import stats


@dataclasses.dataclass
class PhaseCoverageInput:
    """
    Used as input for the multiprocessing task to compute coverage
    """
    time: List[float]
    period: float


class PhaseCoverage:
    @staticmethod
    def phase_coverage(phase_coverage_input: PhaseCoverageInput):
        if phase_coverage_input.period == 0:
            return 1
        else:
            ph = ((phase_coverage_input.time + 0.5 * phase_coverage_input.period) % phase_coverage_input.period -
                  (0.5 * phase_coverage_input.period))
            sph_in = np.sort(ph)
            sph_out = sph_in[::-1]
            sph_in_diff = np.abs(np.diff(sph_in))
            sph_out_diff = np.abs(np.diff(sph_out))
            df = np.min(np.diff(phase_coverage_input.time))
            spaces_in = np.sort(sph_in[np.hstack([*np.argwhere(sph_in_diff > 4 * df).T, len(sph_in) - 1])])
            spaces_out = np.sort(sph_out[np.hstack([*np.argwhere(sph_out_diff > 4 * df).T, len(sph_in) - 1])])
            return np.sum(spaces_in - spaces_out) / phase_coverage_input.period

    @staticmethod
    def compute_phase_coverage(dir: str, time: List[float], period_grid: List[float], cpus=1):
        logging.info("Computing period grid phase coverage")
        period_grid = np.sort(period_grid)
        phase_coverage_inputs = []
        for period in period_grid:
            phase_coverage_inputs.append(PhaseCoverageInput(time, period))
        with Pool(processes=cpus) as pool:
            phase_coverage = np.array(pool.map(PhaseCoverage.phase_coverage, phase_coverage_inputs))
        fig1, (ax1) = plt.subplots(1, 1, figsize=(10, 4), constrained_layout=True)
        fig1.patch.set_facecolor('xkcd:white'),
        bins = len(period_grid) / 100
        phase_coverage = np.array(phase_coverage) * 100
        bin_means, bin_edges, binnumber = stats.binned_statistic(period_grid, phase_coverage, statistic='mean',
                                                                 bins=bins)
        bin_stds, _, _ = stats.binned_statistic(period_grid, phase_coverage, statistic='std', bins=bins)
        bin_width = (bin_edges[1] - bin_edges[0])
        bin_centers = bin_edges[1:] - bin_width / 2
        ax1.plot(period_grid, phase_coverage, color='teal', linestyle='-', alpha=0.15)
        ax1.errorbar(bin_centers, bin_means, yerr=bin_stds / 2, color='peru', marker='.', alpha=0.99, label='-')
        ax1.set_ylabel("Phase Coverage ($\\%$)", fontsize=12)
        ax1.set_xlabel("Period (days)", fontsize=12)
        plt.grid(True)
        plt.savefig(dir + '/phase_coverage.png', dpi=200)





from argparse import ArgumentParser
from sherlockpipe.bayesian_fit.run import run_fit


def fit_parse_args(args=None):
    ap = ArgumentParser(description='Fitting of Sherlock objects of interest')
    ap.add_argument('--object_dir',
                    help="If the object directory is not your current one you need to provide the ABSOLUTE path",
                    required=False)
    ap.add_argument('--candidate', type=str, default=None, help="The CSV candidate signals to be used.", required=False)
    ap.add_argument('--only_initial', dest='only_initial', action='store_true',
                    help="Whether to only run an initial guess of the transit")
    ap.set_defaults(only_initial=False)
    ap.add_argument('--cpus', type=int, default=None, help="The number of CPU cores to be used.", required=False)
    ap.add_argument('--tolerance', type=float, default=0.01, help="The tolerance of the nested sampling algorithm.",
                    required=False)
    ap.add_argument('--mcmc', dest='mcmc', action='store_true', help="Whether to run using mcmc or ns. Default is ns.")
    ap.add_argument('--detrend', dest='detrend', default="hybrid_spline", help="Type of detrending to be used",
                    required=False, choices=['no', 'gp'])
    ap.add_argument('--fit_orbit', dest='fit_orbit', action='store_true', help="Whether to fit eccentricity and "
                                                                               "argument of periastron")
    ap.add_argument('--properties', help="The YAML file to be used as input.", required=False)
    return ap.parse_args(args)


if __name__ == '__main__':
    args = fit_parse_args()
    run_fit(args)



import os
import shutil
import unittest
from lcbuilder.star.starinfo import StarInfo
from lcbuilder.objectinfo.MissionObjectInfo import MissionObjectInfo

from sherlockpipe.scoring.BasicSdeSignalSelector import BasicSdeSignalSelector
from sherlockpipe.scoring.BasicSignalSelector import BasicSignalSelector
from sherlockpipe.scoring.QuorumSdeBorderCorrectedSignalSelector import QuorumSdeBorderCorrectedSignalSelector
from sherlockpipe.scoring.QuorumSnrBorderCorrectedSignalSelector import QuorumSnrBorderCorrectedSignalSelector
from sherlockpipe.scoring.SdeBorderCorrectedSignalSelector import SdeBorderCorrectedSignalSelector
from sherlockpipe.scoring.SnrBorderCorrectedSignalSelector import SnrBorderCorrectedSignalSelector
from sherlockpipe.search.sherlock import Sherlock
from sherlockpipe.search.sherlock_target import SherlockTarget


class TestsSherlock(unittest.TestCase):
    def test_setup_files(self):
        sherlock = Sherlock(None)
        sherlock.setup_files(False, False, False, "inner/")
        self.assertEqual("inner/", sherlock.results_dir)

    def test_setup_detrend(self):
        object_info = MissionObjectInfo('all', "TIC 12345", smooth_enabled=False, high_rms_enabled=False,
                                        high_rms_threshold=3, high_rms_bin_hours=9, auto_detrend_enabled=True,
                                        auto_detrend_ratio=1 / 2, auto_detrend_method="cosine")
        sherlock = Sherlock([SherlockTarget(object_info=object_info, detrends_number=2,
                                            detrend_method="gp", cpu_cores=3)])
        self.assertEqual(False, sherlock.sherlock_targets[0].object_info.smooth_enabled)
        self.assertEqual(False, sherlock.sherlock_targets[0].object_info.high_rms_enabled)
        self.assertEqual(3, sherlock.sherlock_targets[0].object_info.high_rms_threshold)
        self.assertEqual(9, sherlock.sherlock_targets[0].object_info.high_rms_bin_hours)
        self.assertEqual(2, sherlock.sherlock_targets[0].detrends_number)
        self.assertEqual("gp", sherlock.sherlock_targets[0].detrend_method)
        self.assertEqual(3, sherlock.sherlock_targets[0].cpu_cores)
        self.assertEqual(True, sherlock.sherlock_targets[0].object_info.auto_detrend_enabled)
        self.assertEqual(1 / 2, sherlock.sherlock_targets[0].object_info.auto_detrend_ratio)
        self.assertEqual("cosine", sherlock.sherlock_targets[0].object_info.auto_detrend_method)

    def test_setup_transit_adjust_params(self):
        sherlock = Sherlock([SherlockTarget(object_info=None, max_runs=5, period_protect=5, period_min=1,
                                                   period_max=2, bin_minutes=5, cpu_cores=3, snr_min=6, sde_min=5,
                                                   mask_mode="subtract", best_signal_algorithm="quorum",
                                                   quorum_strength=2)])
        self.assertEqual(5, sherlock.sherlock_targets[0].max_runs)
        self.assertEqual(5, sherlock.sherlock_targets[0].period_protect)
        self.assertEqual(1, sherlock.sherlock_targets[0].period_min)
        self.assertEqual(2, sherlock.sherlock_targets[0].period_max)
        self.assertEqual(5, sherlock.sherlock_targets[0].bin_minutes)
        self.assertEqual(3, sherlock.sherlock_targets[0].cpu_cores)
        self.assertEqual(6, sherlock.sherlock_targets[0].snr_min)
        self.assertEqual(5, sherlock.sherlock_targets[0].sde_min)
        self.assertEqual("subtract", sherlock.sherlock_targets[0].mask_mode)
        self.assertEqual("quorum", sherlock.sherlock_targets[0].best_signal_algorithm)
        self.assertEqual(2, sherlock.sherlock_targets[0].quorum_strength)

    def test_scoring_algorithm(self):
        sherlock = Sherlock([SherlockTarget(object_info=None, best_signal_algorithm="basic")])
        self.assertTrue(
            isinstance(sherlock.sherlock_targets[0].signal_score_selectors["basic"], BasicSdeSignalSelector))
        sherlock = Sherlock([SherlockTarget(object_info=None, best_signal_algorithm="border-correct")])
        self.assertTrue(isinstance(sherlock.sherlock_targets[0].signal_score_selectors["border-correct"],
                                   SdeBorderCorrectedSignalSelector))
        sherlock = Sherlock([SherlockTarget(object_info=None, best_signal_algorithm="quorum")])
        self.assertTrue(isinstance(sherlock.sherlock_targets[0].signal_score_selectors["quorum"],
                                   QuorumSdeBorderCorrectedSignalSelector))
        sherlock = Sherlock([SherlockTarget(object_info=None, best_signal_algorithm="basic-snr")])
        self.assertTrue(
            isinstance(sherlock.sherlock_targets[0].signal_score_selectors["basic-snr"], BasicSignalSelector))
        sherlock = Sherlock([SherlockTarget(object_info=None, best_signal_algorithm="border-correct-snr")])
        self.assertTrue(isinstance(sherlock.sherlock_targets[0].signal_score_selectors["border-correct-snr"],
                                   SnrBorderCorrectedSignalSelector))
        sherlock = Sherlock([SherlockTarget(object_info=None, best_signal_algorithm="quorum-snr")])
        self.assertTrue(isinstance(sherlock.sherlock_targets[0].signal_score_selectors["quorum-snr"],
                                   QuorumSnrBorderCorrectedSignalSelector))

    def test_apdate_tois(self):
        sherlock = Sherlock(None)
        sherlock.ois_manager.update_tic_csvs()
        try:
            self.assertTrue(os.path.isfile(sherlock.ois_manager.tois_csv))
        finally:
            os.remove(sherlock.ois_manager.tois_csv)

    def test_apdate_kois(self):
        sherlock = Sherlock(None)
        sherlock.ois_manager.update_kic_csvs()
        try:
            self.assertTrue(os.path.isfile(sherlock.ois_manager.kic_star_csv))
        finally:
            os.remove(sherlock.ois_manager.kic_star_csv)
        try:
            self.assertTrue(os.path.isfile(sherlock.ois_manager.kois_csv))
        finally:
            os.remove(sherlock.ois_manager.kois_csv)

    def test_apdate_epicois(self):
        sherlock = Sherlock(None)
        sherlock.ois_manager.update_epic_csvs()
        try:
            self.assertTrue(os.path.isfile(sherlock.ois_manager.epic_csv))
        finally:
            os.remove(sherlock.ois_manager.epic_csv)

    def test_ois_loaded(self):
        sherlock = Sherlock(None)
        sherlock.load_ois(True, False, False)
        sherlock.filter_hj_ois()
        try:
            self.assertGreater(len(sherlock.ois.index), 100)
            sherlock.limit_ois(0, 5)
            self.assertEqual(len(sherlock.run_ois.index), 5)
            self.assertGreater(len(sherlock.ois.index), 23000)
            self.assertTrue(sherlock.use_ois)
        finally:
            os.remove(sherlock.ois_manager.tois_csv)
            os.remove(sherlock.ois_manager.ctois_csv)
            os.remove(sherlock.ois_manager.kic_star_csv)
            os.remove(sherlock.ois_manager.kois_csv)
            os.remove(sherlock.ois_manager.epic_csv)

    def test_run_empty(self):
        sherlock = Sherlock([])
        self.assertFalse(sherlock.use_ois)
        sherlock.run()
        object_dir = "TIC181084752_[9]"
        self.assertFalse(os.path.exists(object_dir))

    def test_run(self):
        run_dir = "TIC181804752_[9]"
        try:
            Sherlock([SherlockTarget(MissionObjectInfo([9], "TIC 181804752", cadence=1800, smooth_enabled=False,
                                                            high_rms_enabled=False, initial_mask=[[1900, 1901]]),
                                     detrends_number=1, max_runs=1, oversampling=0.05)]).run()
            self.__assert_run_files(run_dir, assert_rms_mask=False)
        finally:
            self.__clean(run_dir)

    def test_run_with_rms_mask(self):
        run_dir = "TIC181804752_[9]"
        try:
            Sherlock([SherlockTarget(MissionObjectInfo([9], "TIC 181804752", cadence=1800, high_rms_enabled=True),
                                     max_runs=1, oversampling=0.05)]).run()
            self.__assert_run_files(run_dir)
        finally:
            self.__clean(run_dir)

    def test_run_with_explore(self):
        run_dir = None
        try:
            Sherlock([SherlockTarget(MissionObjectInfo([9], "TIC 181804752", cadence=1800, high_rms_enabled=True),
                                     detrends_number=1, oversampling=0.05)], True).run()
            run_dir = "TIC181804752_[9]_explore"
            self.assertTrue(os.path.exists(run_dir))
            self.assertTrue(os.path.exists(run_dir + "/Periodogram_Initial_TIC181804752_[9].png"))
            self.assertFalse(os.path.exists(run_dir + "/1"))
        finally:
            self.__clean(run_dir)

    def test_run_with_autodetrend(self):
        run_dir = None
        try:
            Sherlock([SherlockTarget(MissionObjectInfo([5], "TIC 259377017", cadence=1800, auto_detrend_enabled=True),
                                     detrends_number=1, max_runs=1, oversampling=0.05)], True).run()
            run_dir = "TIC259377017_[5]_explore"
            self.assertTrue(os.path.exists(run_dir))
            self.assertTrue(os.path.exists(run_dir + '/Phase_detrend_period_TIC259377017_[5]_8.50_days.png'))
        finally:
            self.__clean(run_dir)

    def test_run_epic_ffi(self):
        run_dir = None
        try:
            Sherlock([SherlockTarget(MissionObjectInfo('all', "EPIC 249631677", cadence=1800, high_rms_enabled=True,
                                                       auto_detrend_enabled=False),
                                     detrends_number=1, max_runs=1, oversampling=0.05)], False).run()
            run_dir = "EPIC249631677_all"
            self.assertTrue(os.path.exists(run_dir))
            self.assertTrue(os.path.exists(run_dir + "/Periodogram_Initial_EPIC249631677_all.png"))
            self.assertTrue(os.path.exists(run_dir + "/1"))
        finally:
            self.__clean(run_dir)

    def test_run_with_star_info(self):
        run_dir = None
        try:
            Sherlock([SherlockTarget(MissionObjectInfo([9], "TIC 181804752", high_rms_enabled=True, cadence=1800,
                                                            star_info=StarInfo(ld_coefficients=(0.15, 0.25),
                                                                               teff=4000,
                                                                               lum=1.50, logg=0.15, radius=0.4,
                                                                               radius_min=0.10, radius_max=0.15,
                                                                               mass=0.3, mass_min=0.05, mass_max=0.075,
                                                                               ra=13.132258, dec=64.334238)),
                                     detrends_number=1, max_runs=1, oversampling=0.05)], True).run()
            run_dir = "TIC181804752_[9]_explore"
            self.assertTrue(os.path.exists(run_dir))
            self.assertTrue(os.path.exists(run_dir + "/Periodogram_Initial_TIC181804752_[9].png"))
            self.assertFalse(os.path.exists(run_dir + "/1"))
            with open(run_dir + '/TIC181804752_[9]_report.log') as f:
                content = f.read()
                self.assertTrue('mass = 0.3' in content)
                self.assertTrue('mass_min = 0.25' in content)
                self.assertTrue('mass_max = 0.375' in content)
                self.assertTrue('radius = 0.4' in content)
                self.assertTrue('radius_min = 0.3' in content)
                self.assertTrue('radius_max = 0.55' in content)
                self.assertTrue('limb-darkening estimates using quadratic LD (a,b)= (0.15, 0.25)' in content)
                self.assertTrue('teff = 4000' in content)
                self.assertTrue('logg = 0.15' in content)
                self.assertTrue('lum = 1.50' in content)
        finally:
            self.__clean(run_dir)

    def test_run_with_transit_customs(self):
        run_dir = None
        try:
            sherlock = Sherlock([SherlockTarget(
                MissionObjectInfo([9], "TIC 181804752", cadence=1800, high_rms_enabled=True),
                detrends_number=1, max_runs=1, oversampling=0.1, t0_fit_margin=0.09,
                duration_grid_step=1.075, fit_method="bls",
                best_signal_algorithm="quorum", quorum_strength=0.31)], False)\
                .run()
            run_dir = "TIC181804752_[9]"
            with open(run_dir + '/TIC181804752_[9]_report.log') as f:
                content = f.read()
                self.assertTrue('Fit method: box' in content)
                self.assertTrue('Duration step: 1.075' in content)
                self.assertTrue('T0 Fit Margin: 0.09' in content)
                self.assertTrue('Oversampling: 0.1' in content)
                self.assertTrue('Signal scoring algorithm: quorum' in content)
                self.assertTrue('Quorum algorithm vote strength: 0.31' in content)
            self.__assert_run_files(run_dir)
        finally:
            self.__clean(run_dir)

    def __assert_run_files(self, object_dir, assert_rms_mask=True):
        run_dir = object_dir + "/1"
        periodogram_file = object_dir + "/Periodogram_Initial_TIC181804752_[9].png"
        periodogram_file1 = object_dir + "/Periodogram_Final_TIC181804752_[9].png"
        rms_mask_file = object_dir + "/rms_mask/High_RMS_Mask_TIC181804752_[9].png"
        lc_file = object_dir + "/lc.csv"
        report_file = object_dir + "/TIC181804752_[9]_report.log"
        candidates_csv_file = object_dir + "/candidates.csv"
        transits_stats_csv_file = object_dir + "/transits_stats.csv"
        try:
            self.assertTrue(os.path.exists(run_dir))
            self.assertTrue(os.path.isfile(periodogram_file))
            self.assertTrue(os.path.isfile(periodogram_file1))
            if assert_rms_mask:
                self.assertTrue(os.path.isfile(rms_mask_file))
            self.assertTrue(os.path.isfile(lc_file))
            self.assertTrue(os.path.isfile(report_file))
            self.assertTrue(os.path.isfile(candidates_csv_file))
            self.assertTrue(os.path.isfile(transits_stats_csv_file))
        finally:
            shutil.rmtree(object_dir, ignore_errors=True)

    def __clean(self, run_dir):
        if run_dir is not None and os.path.isdir(run_dir):
            shutil.rmtree(run_dir, ignore_errors=True)
        mast_download_dir = "mastDownload"
        if mast_download_dir is not None and os.path.isdir(mast_download_dir):
            shutil.rmtree(mast_download_dir, ignore_errors=True)


if __name__ == '__main__':
    unittest.main()



import logging
import math
import multiprocessing
import pickle
import shutil
from typing import List
import pandas
import wotan
import matplotlib.pyplot as plt
import foldedleastsquares as tls
import numpy as np
import os
import sys

from lcbuilder.HarmonicSelector import HarmonicSelector
from lcbuilder.helper import LcbuilderHelper
from lcbuilder.lcbuilder_class import LcBuilder
from lcbuilder.curve_preparer.Flattener import Flattener
from lcbuilder.curve_preparer.Flattener import FlattenInput
from lcbuilder.objectinfo.MissionObjectInfo import MissionObjectInfo
from lcbuilder.objectinfo.InvalidNumberOfSectorsError import InvalidNumberOfSectorsError
from watson.watson import Watson

from sherlockpipe.ois.OisManager import OisManager
from lcbuilder.star.HabitabilityCalculator import HabitabilityCalculator

from sherlockpipe.plot.plotting import save_transit_plot
from sherlockpipe.scoring.helper import compute_border_score, harmonic_spectrum
from sherlockpipe.search.phase_coverage.phase_coverage import PhaseCoverage
from sherlockpipe.search.sherlock_target import SherlockTarget
from sherlockpipe.search.transitresult import TransitResult
from multiprocessing import Pool
from scipy import stats

from sherlockpipe.update import Updater


class Sherlock:
    """
    Main SHERLOCK PIPEline class to be used for loading input, setting up the running parameters and launch the
    analysis of the desired TESS, Kepler, K2 or csv objects light curves.
    """
    MASK_MODES = ['mask', 'subtract']
    RESULTS_DIR = './'
    VALID_DETREND_METHODS = ["biweight", "gp"]
    VALID_PERIODIC_DETREND_METHODS = ["biweight", "gp", "cosine", "cofiam"]
    NUM_CORES = multiprocessing.cpu_count() - 1
    OBJECT_ID_REGEX = "^(KIC|TIC|EPIC)[-_ ]([0-9]+)$"
    NUMBERS_REGEX = "[0-9]+$"
    transits_min_count = {}
    wl_min = {}
    wl_max = {}
    report = {}
    ois = None
    run_ois = None
    config_step = 0
    use_ois = False

    def __init__(self, sherlock_targets: List[SherlockTarget], explore: bool = False, update_ois: bool = False,
                 update_force: bool = False, update_clean: bool = False, cache_dir: str = os.path.expanduser('~') + "/",
                 results_dir: str = None):
        """
        Initializes a Sherlock object, loading the OIs from the csvs, setting up the detrend and transit configurations,
        storing the provided object_infos list and initializing the builders to be used to prepare the light curves for
        the provided object_infos.

        :param bool update_ois: Flag to signal SHERLOCK for updating the TOIs, KOIs and EPICs
        :param List[SherlockTarget] sherlock_targets: a list of objects information to be analysed
        :param bool explore: whether to only run the prepare stage for all objects
        :param bool update_ois: whether ois files should be updated
        :param bool update_force: whether a complete update of metadata should be done
        :param bool update_clean: whether current metadata should be wiped-out before update
        :param str cache_dir: directory to store caches for sherlock.
        :param str results_dir: directory to store results
        """
        self.explore = explore
        self.cache_dir = cache_dir
        self.__setup_logging()
        self.ois_manager = OisManager(self.cache_dir)
        self.setup_files(update_ois, update_force, update_clean, results_dir)
        self.sherlock_targets = sherlock_targets
        self.habitability_calculator = HabitabilityCalculator()
        self.lcbuilder = LcBuilder()

    def setup_files(self, refresh_ois, refresh_force, refresh_clean, results_dir=None):
        """
        Loads the objects of interest data from the downloaded CSVs.

        :param refresh_ois: Flag update the TOIs, KOIs and EPICs
        :param results_dir: Stores the directory to be used for the execution.
        :return: the Sherlock object itself
        """
        self.results_dir = results_dir if results_dir is not None else self.RESULTS_DIR
        self.load_ois(refresh_ois, refresh_force, refresh_clean)
        return self

    def refresh_ois(self):
        """
        Downloads the TOIs, KOIs and EPIC OIs into csv files.

        :return: the Sherlock object itself
        """
        self.ois_manager.update_tic_csvs()
        self.ois_manager.update_kic_csvs()
        self.ois_manager.update_epic_csvs()
        return self

    def load_ois(self, refresh_ois, refresh_force, refresh_clean):
        """
        Loads the csv OIs files into memory

        :return: the Sherlock object itself
        """
        Updater(self.cache_dir).update(refresh_clean, refresh_ois, refresh_force)
        self.ois = self.ois_manager.load_ois()
        return self

    def filter_hj_ois(self):
        """
        Filters the in-memory OIs given some basic filters associated to hot jupiters properties. This method is added
        as an example

        :return: the Sherlock object itself
        """
        self.use_ois = True
        self.run_ois = self.ois[self.ois["Disposition"].notnull()]
        self.run_ois = self.run_ois[self.run_ois["Period (days)"].notnull()]
        self.run_ois = self.run_ois[self.run_ois["Planet Radius (R_Earth)"].notnull()]
        self.run_ois = self.run_ois[self.run_ois["Planet Insolation (Earth Flux)"].notnull()]
        self.run_ois = self.run_ois[self.run_ois["Depth (ppm)"].notnull()]
        self.run_ois = self.run_ois[(self.run_ois["Disposition"] == "KP") | (self.run_ois["Disposition"] == "CP")]
        self.run_ois = self.run_ois[self.run_ois["Period (days)"] < 10]
        self.run_ois = self.run_ois[self.run_ois["Planet Radius (R_Earth)"] > 5]
        self.run_ois = self.run_ois[self.run_ois["Planet Insolation (Earth Flux)"] > 4]
        self.run_ois.sort_values(by=['Object Id', 'OI'])
        return self

    def filter_multiplanet_ois(self):
        """
        Filters the in-memory OIs given some basic filters associated to multiplanet targets. This method is added
        as an example

        :return: the Sherlock object itself
        """
        self.use_ois = True
        self.run_ois = self.ois[self.ois["Disposition"].notnull()]
        self.run_ois = self.run_ois[self.run_ois["Period (days)"].notnull()]
        self.run_ois = self.run_ois[self.run_ois["Period (days)"] > 0]
        self.run_ois = self.run_ois[self.run_ois["Depth (ppm)"].notnull()]
        self.run_ois = self.run_ois[
            (self.run_ois["Disposition"] == "KP") | (self.run_ois["Disposition"] == "CP") | (self.run_ois["Disposition"] == "PC")]
        self.run_ois = self.run_ois[self.run_ois.duplicated(subset=['Object Id'], keep=False)]
        self.run_ois.sort_values(by=['Object Id', 'OI'])
        return self

    def filter_high_snr_long_period_ois(self):
        """
        Filters the in-memory OIs given some basic filters associated to big and long-period targets. This method is added
        as an example

        :return: the Sherlock object itself
        """
        self.use_ois = True
        self.run_ois = self.run_ois[self.run_ois["Disposition"].notnull()]
        self.run_ois = self.run_ois[self.run_ois["Period (days)"].notnull()]
        self.run_ois = self.run_ois[self.run_ois["Period (days)"] > 20]
        self.run_ois = self.run_ois[self.run_ois["Depth (ppm)"].notnull()]
        self.run_ois = self.run_ois[self.run_ois["Depth (ppm)"] > 7500]
        self.run_ois = self.run_ois[
            (self.run_ois["Disposition"] == "KP") | (self.run_ois["Disposition"] == "CP") | (self.run_ois["Disposition"] == "PC")]
        self.run_ois.sort_values(by=['Object Id', 'OI'])
        return self

    def filter_short_period_ois(self):
        """
        Filters the in-memory OIs given some basic filters associated to big and long-period targets. This method is added
        as an example

        :return: the Sherlock object itself
        """
        self.use_ois = True
        self.run_ois = self.ois[self.ois["Period (days)"].notnull()]
        self.run_ois = self.run_ois[self.run_ois["Period (days)"] < 10]
        self.run_ois = self.run_ois.loc[self.run_ois['OI'].str.startswith('TOI', na=False)]
        self.run_ois = self.run_ois[
            (self.run_ois["Disposition"] == "KP") | (self.run_ois["Disposition"] == "CP")]
        self.run_ois = self.run_ois[self.run_ois['Sectors'].apply(lambda x: len(str(x)) <= 2)]
        self.run_ois = self.run_ois.sort_values(by=['Object Id', 'OI'])
        self.run_ois = self.run_ois.drop_duplicates(['Object Id'], keep='first')
        self.run_ois = self.run_ois.sort_values(by=['TESS Mag'])
        self.limit_ois(0, 100)
        return self

    def filter_ois(self, function):
        """Applies a function accepting the Sherlock objects of interests dataframe and stores the result into the
        Sherlock same ois dataframe.

        :param function: the function to be applied to filter the Sherlock OIs.
        :return: the Sherlock object itself
        """
        self.use_ois = True
        self.ois = function(self.ois)
        return self

    def limit_ois(self, offset=0, limit=0):
        """
        Limits the in-memory loaded OIs given an offset and a limit (like a pagination)

        :param offset: the position where the subset must start
        :param limit: maximum number of ois to be returned
        :return: the Sherlock object itself
        """
        if limit == 0:
            limit = len(self.run_ois.index)
        self.run_ois = self.run_ois[offset:limit]
        return self

    def run(self, all_targets_properties: dict = None):
        """
        Entrypoint of Sherlock which launches the main execution for all the input object_infos
        """
        if (self.sherlock_targets is None or len(self.sherlock_targets) == 0) and self.use_ois:
            self.sherlock_targets = [SherlockTarget(MissionObjectInfo('all', object_id))
                                     for object_id in self.run_ois["Object Id"].astype('string').unique()]
        for sherlock_target in self.sherlock_targets:
            if all_targets_properties is not None:
                for key, value in all_targets_properties.items():
                    if '.' in key:
                        keys = key.split('.')
                        object = getattr(sherlock_target, keys[0])
                        key = keys[1]
                        setattr(object, key, value)
                    else:
                        setattr(sherlock_target, key, value)
            self.__run_object(sherlock_target)

    def __min_transits_count(self, lc_build, sherlock_target):
        return sherlock_target.min_transits_count \
                if sherlock_target.min_transits_count > 0 else lc_build.transits_min_count

    def __run_object(self, sherlock_target: SherlockTarget):
        """
        Performs the analysis for one object_info

        :param sherlock_target: The object to be analysed.
        """
        sherlock_id = sherlock_target.object_info.sherlock_id()
        mission_id = sherlock_target.object_info.mission_id()
        try:
            lc_build = self.__prepare(sherlock_target)
            object_dir = self.__init_object_dir(sherlock_id)
            if lc_build.lc_data is not None:
                lc_build.lc_data.to_csv(object_dir + "/lc_data.csv")
            time = lc_build.lc.time.value
            flux = lc_build.lc.flux.value
            flux_err = lc_build.lc.flux_err.value
            min_transits_count = self.__min_transits_count(lc_build, sherlock_target)
            period_grid, oversampling = LcbuilderHelper.calculate_period_grid(time, sherlock_target.period_min,
                                                                sherlock_target.period_max,
                                                                sherlock_target.oversampling,
                                                                lc_build.star_info,
                                                                1)
            id_run = 1
            best_signal_score = 1
            self.report[sherlock_id] = []
            logging.info('================================================')
            logging.info('SEARCH RUNS with period grid: [%.2f - %.2f] and length %.0f', np.min(period_grid),
                         np.max(period_grid), len(period_grid))
            logging.info('================================================')
            object_dir = self.__init_object_dir(sherlock_target.object_info.sherlock_id())
            if sherlock_target.compute_phase_coverage:
                PhaseCoverage.compute_phase_coverage(object_dir, time, period_grid,
                                                     sherlock_target.cpu_cores)
            lcs, wl = self.__detrend(sherlock_target, time, flux, flux_err,
                                     lc_build.star_info)
            lcs = np.concatenate(([flux], lcs), axis=0)
            object_info = sherlock_target.object_info
            i = 0
            for lc in lcs:
                lc_df = pandas.DataFrame(columns=['#time', 'flux', 'flux_err'])
                args = np.argwhere(~np.isnan(lc)).flatten()
                lc_df['#time'] = time[args]
                lc_df['flux'] = lc[args]
                lc_df['flux_err'] = np.array(flux_err[args])
                flux_err_mean = np.nanmean(lc_df['flux_err'])
                lc_df.loc[(lc_df['flux_err'] == 0) | np.isnan(lc_df['flux_err']), 'flux_err'] = flux_err_mean
                lc_df.to_csv(object_dir + "/lc_" + str(i) + ".csv", index=False)
                i = i + 1
            wl = np.concatenate(([0], wl), axis=0)
            transits_stats_df = pandas.DataFrame(columns=['candidate', 't0', 'depth', 'depth_err'])
            while not self.explore and best_signal_score == 1 and id_run <= sherlock_target.max_runs:
                self.__setup_object_logging(sherlock_id, False)
                object_report = {}
                logging.info("________________________________ run %s________________________________", id_run)
                transit_results, signal_selection = \
                    self.__analyse(sherlock_target, time, lcs, flux_err, lc_build.star_info, id_run,
                                   min_transits_count, lc_build.cadence, self.report[sherlock_id], wl,
                                   period_grid, lc_build.detrend_period)
                all_nan_results = len(np.argwhere(~np.isnan(signal_selection.transit_result.t0s)).flatten()) == 0
                if not all_nan_results:
                    for index in np.arange(len(signal_selection.transit_result.t0s)):
                        transits_stats_df = transits_stats_df.append({'candidate': str(int(id_run - 1)),
                                                  't0': signal_selection.transit_result.t0s[index],
                                                  'depth': signal_selection.transit_result.depths[index],
                                                  'depth_err': signal_selection.transit_result.depths_err[index]},
                                                 ignore_index=True)
                transits_stats_df = transits_stats_df.sort_values(by=['candidate', 't0'], ascending=True)
                transits_stats_df.to_csv(object_dir + "transits_stats.csv", index=False)
                best_signal_score = signal_selection.score
                object_report["Object Id"] = mission_id
                object_report["run"] = id_run
                object_report["score"] = best_signal_score
                object_report["curve"] = str(signal_selection.curve_index)
                object_report["snr"] = signal_selection.transit_result.snr
                object_report["sde"] = signal_selection.transit_result.sde
                object_report["fap"] = signal_selection.transit_result.fap
                object_report["border_score"] = signal_selection.transit_result.border_score
                object_report["harmonic"] = signal_selection.transit_result.harmonic
                object_report["period"] = signal_selection.transit_result.period
                object_report["per_err"] = signal_selection.transit_result.per_err
                object_report["duration"] = signal_selection.transit_result.duration * 60 * 24
                object_report["t0"] = signal_selection.transit_result.t0
                object_report["depth"] = signal_selection.transit_result.depth
                object_report["depth_err"] = signal_selection.transit_result.depth_err
                object_report["depth_sig"] = signal_selection.transit_result.depth / signal_selection.transit_result.depth_err
                if signal_selection.transit_result is not None and signal_selection.transit_result.results is not None\
                        and not all_nan_results:
                    object_report['rp_rs'] = signal_selection.transit_result.results.rp_rs
                    real_transit_args = np.argwhere(~np.isnan(signal_selection.transit_result
                                                              .results.transit_depths))
                    object_report["transits"] = np.array(signal_selection.transit_result
                                                         .results.transit_times)[real_transit_args.flatten()]
                    object_report["transits"] = ','.join(map(str, object_report["transits"]))
                else:
                    object_report['rp_rs'] = 0
                    object_report["transits"] = ""
                object_report["sectors"] = ','.join(map(str, lc_build.sectors)) \
                    if lc_build.sectors is not None and len(lc_build.sectors) > 0 else None
                object_report["oi"] = self.__find_matching_oi(sherlock_target.object_info, object_report["period"],
                                                              object_report['t0'])
                if best_signal_score == 1:
                    logging.info('New best signal is good enough to keep searching. Going to the next run.')
                    time, lcs = self.__apply_mask_from_transit_results(sherlock_target, time, lcs, transit_results,
                                                                        signal_selection.curve_index)
                    id_run += 1
                    if id_run > sherlock_target.max_runs:
                        logging.info("Max runs limit of %.0f is reached. Stopping.", sherlock_target.max_runs)
                else:
                    logging.info('New best signal does not look very promising. End')
                self.report[sherlock_id].append(object_report)
                self.__setup_object_report_logging(sherlock_id, True)
                logging.info("Listing most promising candidates for ID %s:", sherlock_id)
                logging.info("%-12s%-10s%-10s%-10s%-8s%-8s%-11s%-11s%-8s%-8s%-14s%-14s%-12s%-25s%-10s%-18s%-20s",
                             "Detrend no.", "Period", "Per_err", "Duration", "T0", "Depth", "Depth_err", "Depth_sig",
                             "SNR", "SDE", "Border_score", "Matching OI", "Harmonic",
                             "Planet radius (R_Earth)", "Rp/Rs", "Semi-major axis", "Habitability Zone")
                if sherlock_id in self.report:
                    candidates_df = pandas.DataFrame(columns=['curve', 'period', 'per_err', 'duration', 't0', 'depth',
                                                              'snr', 'sde', 'border_score', 'oi', 'rad_p', 'rp_rs',
                                                              'a', 'hz'])
                    i = 1
                    for report in self.report[sherlock_id]:
                        a, habitability_zone = self.habitability_calculator\
                            .calculate_hz_score(lc_build.star_info.teff, lc_build.star_info.mass,
                                                lc_build.star_info.lum, report["period"])
                        report['a'] = a
                        report['hz'] = habitability_zone
                        if lc_build.star_info.radius_assumed:
                            report['rad_p'] = np.nan
                        else:
                            report['rad_p'] = self.__calculate_planet_radius(lc_build.star_info, report["depth"])
                        logging.info("%-12s%-10.4f%-10.5f%-10.2f%-8.2f%-8.3f%-11.3f%-11.3f%-8.2f%-8.2f%-14.2f%-14s%-12s%-25.5f%-10.5f%-18.5f%-20s",
                                     report["curve"], report["period"], report["per_err"],
                                     report["duration"], report["t0"], report["depth"], report["depth_err"],
                                     report["depth_sig"], report["snr"], report["sde"],
                                     report["border_score"], report["oi"], report["harmonic"],
                                     report['rad_p'], report['rp_rs'], a, habitability_zone)
                        candidates_df = candidates_df.append(report, ignore_index=True)
                        i = i + 1
                    candidates_df.to_csv(object_dir + "candidates.csv", index=False)
        except InvalidNumberOfSectorsError:
            logging.exception("Invalid number of sectors exception")
            self.__remove_object_dir(sherlock_id)
        except Exception as e:
            logging.exception("Unexpected exception")

    def noise(self, time, flux, signal_power):
        from scipy.signal import periodogram, welch
        f_p, psd_p = periodogram(flux)
        f_w, psd_w = welch(flux)
        power_p = np.trapz(psd_p, f_p)
        power_w = np.trapz(psd_w, f_w)
        snr_p = signal_power / power_p
        snr_w = signal_power / power_w
        return snr_p

    def __init_object_dir(self, object_id, clean_dir=False):
        dir = self.results_dir + str(object_id)
        dir = dir.replace(" ", "_")
        dir = dir if not self.explore else dir + '_explore'
        if clean_dir:
            self.__remove_object_dir(object_id)
        if not os.path.exists(dir):
            os.makedirs(dir)
        return dir + "/"

    def __remove_object_dir(self, object_id):
        dir = self.results_dir + str(object_id)
        dir = dir.replace(" ", "_")
        if os.path.exists(dir) and os.path.isdir(dir):
            shutil.rmtree(dir, ignore_errors=True)

    def __init_object_run_dir(self, object_id, run_no, clean_dir=False):
        dir = self.results_dir + str(object_id) + "/" + str(run_no)
        dir = dir.replace(" ", "_")
        if clean_dir and os.path.exists(dir) and os.path.isdir(dir):
            shutil.rmtree(dir, ignore_errors=True)
        if not os.path.exists(dir):
            os.makedirs(dir)
        return dir + "/"

    def __setup_logging(self):
        if not isinstance(logging.root, logging.RootLogger):
            logging.root = logging.RootLogger(logging.INFO)
        logging.captureWarnings(True)
        formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
        logger = logging.getLogger()
        while len(logger.handlers) > 0:
            logger.handlers.pop()
        logger.setLevel(logging.INFO)
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.INFO)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    def __setup_object_logging(self, object_id, clean_dir=True):
        if not isinstance(logging.root, logging.RootLogger):
            logging.root = logging.RootLogger(logging.INFO)
        object_dir = self.__init_object_dir(object_id, clean_dir)
        logger = logging.getLogger()
        while len(logger.handlers) > 1:
            logger.handlers.pop()
        formatter = logging.Formatter()
        handler = logging.FileHandler(object_dir + str(object_id) + "_report.log")
        handler.setLevel(logging.INFO)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return object_dir

    def __setup_object_report_logging(self, object_id, clean=False):
        if not isinstance(logging.root, logging.RootLogger):
            logging.root = logging.RootLogger(logging.INFO)
        object_dir = self.__setup_object_logging(object_id, False)
        logger = logging.getLogger()
        logger.handlers.pop()
        formatter = logging.Formatter()
        file = object_dir + str(object_id) + "_candidates.log"
        if clean and os.path.exists(file):
            os.remove(file)
        handler = logging.FileHandler(file)
        handler.setLevel(logging.INFO)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    def __prepare(self, sherlock_target):
        object_info = sherlock_target.object_info
        sherlock_id = object_info.sherlock_id()
        object_dir = self.__setup_object_logging(sherlock_id)
        logging.info('SHERLOCK (Searching for Hints of Exoplanets fRom Lightcurves Of spaCe-base seeKers)')
        logging.info('Version %s', sys.modules["sherlockpipe"].__version__)
        logging.info('MODE: %s', "ANALYSE" if not self.explore else "EXPLORE")
        logging.info('ID: %s', sherlock_id)
        if sherlock_target.source_properties_file is not None:
            logging.info("Storing properties file %s", sherlock_target.source_properties_file)
            shutil.copy(sherlock_target.source_properties_file, object_dir + "/properties.yaml")
        logging.info('================================================')
        logging.info('USER DEFINITIONS')
        logging.info('================================================')
        if sherlock_target.detrend_method == "gp":
            logging.info('Detrend method: Gaussian Process Matern 2/3')
        else:
            logging.info('Detrend method: Bi-Weight')
        logging.info('No of detrend models applied: %s', sherlock_target.detrends_number)
        logging.info('Period planet protected: %.1f', sherlock_target.period_protect)
        logging.info('Minimum Period (d): %.1f', sherlock_target.period_min)
        logging.info('Maximum Period (d): %.1f', sherlock_target.period_max)
        logging.info('Binning size (min): %.1f', sherlock_target.bin_minutes)
        if object_info.initial_mask is not None:
            logging.info('Mask: yes')
        else:
            logging.info('Mask: no')
        if object_info.initial_transit_mask is not None:
            logging.info('Transit Mask: yes')
        else:
            logging.info('Transit Mask: no')
        logging.info('Threshold limit for SNR: %.1f', sherlock_target.snr_min)
        logging.info('Threshold limit for SDE: %.1f', sherlock_target.sde_min)
        if object_info.outliers_sigma:
            logging.info('Sigma threshold for upper outliers clipping: %.1f', object_info.outliers_sigma)
        if object_info.lower_outliers_sigma:
            logging.info('Sigma threshold for lower outliers clipping: %.1f', object_info.lower_outliers_sigma)
        logging.info("Fit method: %s", sherlock_target.fit_method)
        if sherlock_target.oversampling is not None:
            logging.info('Oversampling: %.3f', sherlock_target.oversampling)
        if sherlock_target.duration_grid_step is not None:
            logging.info('Duration step: %.3f', sherlock_target.duration_grid_step)
        if sherlock_target.t0_fit_margin is not None:
            logging.info('T0 Fit Margin: %.3f', sherlock_target.t0_fit_margin)
        logging.info('Signal scoring algorithm: %s', sherlock_target.best_signal_algorithm)
        if sherlock_target.best_signal_algorithm == sherlock_target.VALID_SIGNAL_SELECTORS[2]:
            logging.info('Quorum algorithm vote strength: %.2f',
                         sherlock_target.signal_score_selectors[sherlock_target.VALID_SIGNAL_SELECTORS[2]].strength)
        elif sherlock_target.best_signal_algorithm == sherlock_target.VALID_SIGNAL_SELECTORS[5]:
            logging.info('Quorum algorithm vote strength: %.2f',
                         sherlock_target.signal_score_selectors[sherlock_target.VALID_SIGNAL_SELECTORS[5]].strength)
        mission, mission_prefix, object_num = self.lcbuilder.parse_object_info(object_info.mission_id())
        if object_info.reduce_simple_oscillations and \
                object_info.oscillation_max_period < object_info.oscillation_min_period:
            logging.info("Stellar oscillation period has been set to empty. Defaulting to 1/3 the minimum search period")
            object_info.oscillation_max_period = sherlock_target.period_min / 3
        lc_build = self.lcbuilder.build(object_info, object_dir, cpus=sherlock_target.cpu_cores)
        min_transits_count = self.__min_transits_count(lc_build, sherlock_target)
        logging.info('Minimum number of transits: %s', min_transits_count)
        lightcurve_timespan = lc_build.lc.time[len(lc_build.lc.time) - 1] - lc_build.lc.time[0]
        if sherlock_target.search_zone is not None and not (lc_build.star_info.mass_assumed or
                                                            lc_build.star_info.radius_assumed):
            logging.info("Selected search zone: %s. Minimum and maximum periods will be calculated.",
                         sherlock_target.search_zone)
            min_and_max_period = sherlock_target.search_zones_resolvers[
                sherlock_target.search_zone].calculate_period_range(lc_build.star_info)
            if min_and_max_period is None:
                logging.info("Selected search zone was %s but cannot be calculated. Defaulting to minimum and " +
                             "maximum input periods", sherlock_target.search_zone)
            else:
                logging.info("Selected search zone periods are [%.2f, %.2f] days", min_and_max_period[0],
                             min_and_max_period[1])
                if min_and_max_period[0] > lightcurve_timespan or min_and_max_period[1] > lightcurve_timespan:
                    logging.info("Selected search zone period values are greater than lightcurve dataset. " +
                                 "Defaulting to minimum and maximum input periods.")
                else:
                    sherlock_target.period_min = min_and_max_period[0]
                    sherlock_target.period_max = min_and_max_period[1]
        elif sherlock_target.search_zone is not None:
            logging.info("Selected search zone was %s but star catalog info was not found or wasn't complete. " +
                         "Defaulting to minimum and maximum input periods.", sherlock_target.search_zone)
        if lc_build.sectors is not None:
            sectors_count = len(lc_build.sectors)
            logging.info('================================================')
            logging.info('SECTORS/QUARTERS/CAMPAIGNS INFO')
            logging.info('================================================')
            logging.info('Sectors/Quarters/Campaigns : %s', lc_build.sectors)
            if sectors_count < sherlock_target.min_sectors or sectors_count > sherlock_target.max_sectors:
                raise InvalidNumberOfSectorsError("The object " + sherlock_id + " contains " + str(sectors_count) +
                                                  " sectors/quarters/campaigns and the min and max selected are [" +
                                                  str(sherlock_target.min_sectors) + ", " + str(
                    sherlock_target.max_sectors) + "].")
        logging.info('================================================')
        logging.info("Detrend Window length / Kernel size")
        logging.info('================================================')
        if sherlock_target.detrend_l_min is not None and sherlock_target.detrend_l_max is not None:
            logging.info("Using user input WL / KS")
            self.wl_min[sherlock_id] = sherlock_target.detrend_l_min
            self.wl_max[sherlock_id] = sherlock_target.detrend_l_max
        else:
            logging.info("Using transit duration based WL or fixed KS")
            min_transit_duration = wotan.t14(R_s=lc_build.star_info.radius, M_s=lc_build.star_info.mass,
                                             P=sherlock_target.period_min, small_planet=True)
            max_transit_duration = wotan.t14(R_s=lc_build.star_info.radius, M_s=lc_build.star_info.mass,
                                             P=sherlock_target.period_max, small_planet=False)
            self.wl_min[sherlock_id] = 4 * min_transit_duration
            self.wl_max[sherlock_id] = 8 * max_transit_duration
        logging.info("wl/ks_min: %.2f", self.wl_min[sherlock_id])
        logging.info("wl/ks_max: %.2f", self.wl_max[sherlock_id])
        logging.info('================================================')
        if not sherlock_target.period_min:
            sherlock_target.period_min = lc_build.detrend_period * 4
            logging.info("Setting Min Period to %.3f due to auto_detrend period", sherlock_target.period_min)
        if lc_build.sectors is not None:
            logging.info("======================================")
            logging.info("Field Of View Plots")
            logging.info("======================================")
            fov_dir = object_dir + "/fov"
            Watson.vetting_field_of_view(fov_dir, mission, object_num, lc_build.cadence, lc_build.star_info.ra,
                                         lc_build.star_info.dec, lc_build.sectors if isinstance(lc_build.sectors, list)
                                         else lc_build.sectors.tolist(), lc_build.tpf_source, lc_build.tpf_apertures)
        if sherlock_target.ois_mask and self.ois is not None:
            logging.info("Masking OIS")
            target_ois = self.ois[self.ois["Object Id"] == object_info.mission_id()]
            target_ois = target_ois[target_ois["OI"].notnull()]
            transits_dict = []
            for index, oi_row in target_ois.iterrows():
                epoch = oi_row['Epoch (BJD)']
                epoch = LcbuilderHelper.correct_epoch(mission, epoch)
                transits_dict.append({'P': oi_row["Period (days)"], 'T0': epoch,
                                      'D': oi_row['Duration (hours)'] * 60 * 2})
            lc_build.lc = LcbuilderHelper.mask_transits_dict(lc_build.lc, transits_dict)
        return lc_build

    def __analyse(self, sherlock_target, time, lcs, flux_err, star_info, id_run, transits_min_count, cadence, report,
                  wl, period_grid, detrend_source_period):
        logging.info('=================================')
        logging.info('SEARCH OF SIGNALS - Run %s', id_run)
        logging.info('=================================')
        transit_results = self.__identify_signals(sherlock_target, time, lcs, flux_err, star_info, transits_min_count,
                                                  wl, id_run, cadence, report, period_grid, detrend_source_period)
        run_dir = self.__init_object_dir(star_info.object_id) + '/' + str(id_run) + '/'
        signal_selection = sherlock_target.signal_score_selectors[sherlock_target.best_signal_algorithm]\
            .select(id_run, sherlock_target, star_info, transits_min_count, time, lcs, transit_results, wl, cadence)
        title = 'Run ' + str(id_run) + '# curve: SELECTED' + \
                ' # P=' + format(signal_selection.transit_result.period, '.2f') + 'd # T0=' + \
                format(signal_selection.transit_result.t0, '.2f') + ' # Depth=' + \
                format(signal_selection.transit_result.depth, '.4f') + "ppt # Dur=" + \
                format(signal_selection.transit_result.duration * 24 * 60, '.0f') + 'm # SNR:' + \
                str(format(signal_selection.transit_result.snr, '.2f')) + ' # SDE:' + \
                str(format(signal_selection.transit_result.sde, '.2f')) + ' # BS:' + \
                str(format(signal_selection.transit_result.border_score, '.2f'))
        file = 'Run_' + str(id_run) + '_SELECTED_' + str(signal_selection.curve_index) + '_' + str(star_info.object_id) + '.png'
        save_transit_plot(star_info.object_id, title, run_dir, file, time, lcs[signal_selection.curve_index],
                          signal_selection.transit_result, cadence, id_run, sherlock_target.use_harmonics_spectra)
        if sherlock_target.pickle_mode == 'all':
            with open(run_dir + 'search_results.pickle', 'wb') as search_results_file:
                pickle.dump(transit_results, search_results_file, protocol=pickle.HIGHEST_PROTOCOL)
        elif sherlock_target.pickle_mode == 'selected':
            with open(run_dir + 'search_results.pickle', 'wb') as search_results_file:
                pickle.dump(transit_results[signal_selection.curve_index], search_results_file,
                            protocol=pickle.HIGHEST_PROTOCOL)
        logging.info(signal_selection.get_message())
        return transit_results, signal_selection

    def __detrend(self, sherlock_target, time, lc, flux_err, star_info):
        wl_min = self.wl_min[star_info.object_id]
        wl_max = self.wl_max[star_info.object_id]
        bins = len(time) * 2 / sherlock_target.bin_minutes
        bin_means, bin_edges, binnumber = stats.binned_statistic(time, lc, statistic='mean', bins=bins)
        logging.info('=================================')
        logging.info('MODELS IN THE DETRENDING')
        logging.info('=================================')
        logging.info("%-25s%-17s%-15s%-11s%-15s", "light_curve", "Detrend_method", "win/ker_size", "RMS (ppm)",
                     "RMS_10min (ppm)")
        logging.info("%-25s%-17s%-15s%-11.2f%-15.2f", "PDCSAP_FLUX", "---", "---", np.std(lc) * 1e6,
                     np.std(bin_means[~np.isnan(bin_means)]) * 1e6)
        wl_step = (wl_max - wl_min) / sherlock_target.detrends_number
        wl = np.arange(wl_min, wl_max, wl_step)  # we define all the posibles window_length that we apply
        final_lcs = np.zeros((len(wl), len(lc)))
        flatten_inputs = []
        flattener = Flattener()
        if sherlock_target.detrend_cores > 1:
            for i in range(0, len(wl)):
                flatten_inputs.append(FlattenInput(time, lc, wl[i], sherlock_target.bin_minutes))
            if sherlock_target.detrend_method == 'gp':
                flatten_results = self.run_multiprocessing(sherlock_target.cpu_cores, flattener.flatten_gp, flatten_inputs)
            else:
                flatten_results = self.run_multiprocessing(sherlock_target.cpu_cores, flattener.flatten_bw, flatten_inputs)
        else:
            flatten_results = []
            for i in range(0, len(wl)):
                if sherlock_target.detrend_method == 'gp':
                    flatten_results.append(flattener.flatten_gp(FlattenInput(time, lc, wl[i], sherlock_target.bin_minutes)))
                else:
                    flatten_results.append(flattener.flatten_bw(FlattenInput(time, lc, wl[i], sherlock_target.bin_minutes)))
        i = 0
        # Plot entire curve detrends
        for flatten_lc_detrended, lc_trend, bin_centers, bin_means, flatten_wl in flatten_results:
            final_lcs[i] = flatten_lc_detrended
            logging.info("%-25s%-17s%-15.4f%-11.2f%-15.2f", 'flatten_lc & trend_lc ' + str(i),
                         sherlock_target.detrend_method,
                         flatten_wl, np.std(flatten_lc_detrended) * 1e6, np.std(bin_means[~np.isnan(bin_means)]) * 1e6)
            self.__plot_detrends(star_info.object_id, sherlock_target.detrends_number,
                                 sherlock_target.detrend_method, wl, time, lc, flatten_results, 0, len(time) - 1,
                                 'Detrends_' + str(star_info.object_id) + '.png')
            i = i + 1
        # Plot tokenized curve detrends
        dif = time[1:] - time[:-1]
        jumps = np.where(dif > 3)[0]
        jumps = np.append(jumps, len(time))
        previous_jump_index = 0
        for jumpIndex in jumps:
            self.__plot_detrends(star_info.object_id, sherlock_target.detrends_number, sherlock_target.detrend_method,
                                 wl, time, lc, flatten_results, previous_jump_index, jumpIndex)
            previous_jump_index = jumpIndex
        return final_lcs, wl

    def __plot_detrends(self, object_id, detrends_number, detrend_method, window_lengths, time, original_lc,
                        flatten_results, low_index, high_index, filename=None):
        plot_dir = self.__init_object_dir(object_id) + "/detrends/"
        if not os.path.exists(plot_dir):
            os.mkdir(plot_dir)
        figsize = (12, 4 * detrends_number)  # x,y
        rows = detrends_number
        cols = 1
        min_original = np.min(original_lc[low_index:high_index])
        shift = 2 * (1.0 - (np.min(original_lc)))  # shift in the between the raw and detrended data
        fig, axs = plt.subplots(rows, cols, figsize=figsize, constrained_layout=True)
        if detrends_number > 1:
            axs = self.__trim_axs(axs, len(window_lengths))
        plot_axs = axs
        i = 0
        time_partial = time[low_index:high_index]
        original_lc_partial = original_lc[low_index:high_index]
        for flatten_lc_detrended, lc_trend, bin_centers, bin_means, flatten_wl in flatten_results:
            lc_trend_partial = lc_trend[low_index:high_index]
            dist = 1.3 * ((1 - min_original) + (np.max(flatten_lc_detrended) - 1))
            if detrends_number > 1:
                plot_axs = axs[i]
            if detrend_method == 'gp':
                plot_axs.set_title('ks=%s' % str(np.around(flatten_wl, decimals=4)))
            else:
                plot_axs.set_title('ws=%s d' % str(np.around(flatten_wl, decimals=4)))
            plot_axs.scatter(time_partial[1:], original_lc_partial[1:], color='black', s=1, alpha=0.25, rasterized=True)
            plot_axs.plot(time_partial[1:], lc_trend_partial[1:], linewidth=2, color='firebrick', alpha=1.)
            plot_axs.scatter(time_partial[1:], flatten_lc_detrended[low_index:high_index][1:] - dist, color='black', s=1, alpha=0.75, rasterized=True)
            plot_axs.set_ylabel("Flux norm.")
            plot_axs.set_xlabel("Time (d)")
            i = i + 1
        filename = filename if filename is not None else 'Detrends_' + str(object_id) + '_time_' + \
                                                         str(time_partial[1]) + '_' + str(time_partial[-1]) + '.png'
        plt.savefig(plot_dir + filename, dpi=200)
        fig.clf()
        plt.close(fig)

    def __identify_signals(self, sherlock_target, time, lcs, flux_err, star_info, transits_min_count, wl, id_run,
                           cadence, report, period_grid, detrend_source_period):
        object_info = sherlock_target.object_info
        detrend_logging_customs = 'ker_size' if sherlock_target.detrend_method == 'gp' else "win_size"
        logging.info("%-12s%-12s%-10s%-8s%-18s%-12s%-12s%-14s%-14s%-12s%-12s%-16s%-14s%-12s%-25s%-10s%-18s%-20s",
                     detrend_logging_customs, "Period", "Per_err", "N.Tran", "Mean Depth (ppt)", 'Depth_err',
                     'Depth_sig', "T. dur (min)", "T0", "SNR", "SDE", "Border_score", "Matching OI", "Harmonic",
                     "Planet radius (R_Earth)", "Rp/Rs", "Semi-major axis", "Habitability Zone")
        transit_results = {}
        plot_dir = self.__init_object_run_dir(star_info.object_id, id_run)
        if not sherlock_target.ignore_original:
            transit_result = self.__adjust_transit(sherlock_target, time, lcs[0], star_info, transits_min_count,
                                                   transit_results, report, cadence, period_grid, detrend_source_period)
            r_planet = self.__calculate_planet_radius(star_info, transit_result.depth)
            rp_rs = transit_result.results.rp_rs
            a, habitability_zone = self.habitability_calculator \
                .calculate_hz_score(star_info.teff, star_info.mass, star_info.lum, transit_result.period)
            oi = self.__find_matching_oi(object_info, transit_result.period, transit_result.t0)
            logging.info('%-12s%-12.5f%-10.6f%-8s%-18.3f%-12.3f%-12.3f%-14.1f%-14.4f%-12.3f%-12.3f%-16.2f%-14s%-12s%-25.5f%-10.5f%-18.5f%-20s',
                         "PDCSAP_FLUX", transit_result.period,
                         transit_result.per_err, transit_result.count, transit_result.depth, transit_result.depth_err,
                         transit_result.depth / transit_result.depth_err,
                         transit_result.duration * 24 * 60, transit_result.t0, transit_result.snr, transit_result.sde,
                         transit_result.border_score, oi, transit_result.harmonic, r_planet, rp_rs, a,
                         habitability_zone)
            plot_title = 'Run ' + str(id_run) + 'PDCSAP_FLUX # P=' + \
                         format(transit_result.period, '.2f') + 'd # T0=' + format(transit_result.t0, '.2f') + \
                         ' # Depth=' + format(transit_result.depth, '.4f') + 'ppt # Dur=' + \
                         format(transit_result.duration * 24 * 60, '.0f') + 'm # SNR:' + \
                         str(format(transit_result.snr, '.2f')) + ' # SDE:' + str(format(transit_result.sde, '.2f')) + \
                         ' # BS:' + str(format(transit_result.border_score, '.2f'))
            plot_file = 'Run_' + str(id_run) + '_PDCSAP-FLUX_' + str(star_info.object_id) + '.png'
            save_transit_plot(star_info.object_id, plot_title, plot_dir, plot_file, time, lcs[0], transit_result,
                              cadence, id_run, sherlock_target.use_harmonics_spectra)
        else:
            transit_result = TransitResult(None, None, 0, 0, 0, 0, [], [], 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, [])
        transit_results[0] = transit_result
        for i in range(1, len(wl)):
            transit_result = self.__adjust_transit(sherlock_target, time, lcs[i], star_info, transits_min_count,
                                                   transit_results, report, cadence, period_grid, detrend_source_period)
            transit_results[i] = transit_result
            r_planet = self.__calculate_planet_radius(star_info, transit_result.depth)
            rp_rs = transit_result.results.rp_rs
            a, habitability_zone = self.habitability_calculator \
                .calculate_hz_score(star_info.teff, star_info.mass, star_info.lum, transit_result.period)
            oi = self.__find_matching_oi(object_info, transit_result.period, transit_result.t0)
            logging.info('%-12.4f%-12.5f%-10.6f%-8s%-18.3f%-12.3f%-12.3f%-14.1f%-14.4f%-12.3f%-12.3f%-16.2f%-14s%-12s%-25.5f%-10.5f%-18.5f%-20s',
                         wl[i], transit_result.period,
                     transit_result.per_err, transit_result.count, transit_result.depth, transit_result.depth_err,
                         transit_result.depth / transit_result.depth_err,
                     transit_result.duration * 24 * 60, transit_result.t0, transit_result.snr, transit_result.sde,
                         transit_result.border_score, oi, transit_result.harmonic, r_planet, rp_rs, a,
                     habitability_zone)
            detrend_file_title_customs = 'ker_size' if sherlock_target.detrend_method == 'gp' else 'win_size'
            detrend_file_name_customs = 'ks' if sherlock_target.detrend_method == 'gp' else 'ws'
            title = 'Run ' + str(id_run) + '# ' + detrend_file_title_customs + ':' + str(format(wl[i], '.4f')) + \
                    ' # P=' + format(transit_result.period, '.2f') + 'd # T0=' + \
                    format(transit_result.t0, '.2f') + ' # Depth=' + format(transit_result.depth, '.4f') + "ppt # Dur=" + \
                    format(transit_result.duration * 24 * 60, '.0f') + 'm # SNR:' + \
                    str(format(transit_result.snr, '.2f')) + ' # SDE:' + str(format(transit_result.sde, '.2f')) + \
                    ' # BS:' + str(format(transit_result.border_score, '.2f'))
            file = 'Run_' + str(id_run) + '_' + detrend_file_name_customs + '=' + str(format(wl[i], '.4f')) + '_' + \
                   str(star_info.object_id) + '.png'
            save_transit_plot(star_info.object_id, title, plot_dir, file, time, lcs[i], transit_result, cadence,
                              id_run, sherlock_target.use_harmonics_spectra)
        return transit_results

    def __find_matching_oi(self, object_info, period, epoch):
        oi = ""
        if self.ois is not None:
            mission, mission_id, target_id = LcBuilder().parse_object_info(object_info.mission_id())
            corrected_epoch = LcbuilderHelper.normalize_mission_epoch(mission, epoch)
            existing_period_in_object = self.ois[(self.ois["Object Id"] == object_info.mission_id())]
            exists_oi = False
            for index, row in existing_period_in_object.iterrows():
                if HarmonicSelector.is_harmonic(0, (np.abs(corrected_epoch - row['Epoch (BJD)']) % row["Period (days)"]), period, row["Period (days)"]):
                    oi = row["OI"]
                    break
        return oi

    def __adjust_transit(self, sherlock_target, time, lc, star_info, transits_min_count, run_results, report, cadence,
                         period_grid, detrend_source_period):
        oversampling = sherlock_target.oversampling
        model = tls.transitleastsquares(time, lc)
        power_args = {"transit_template": sherlock_target.fit_method, "period_min": sherlock_target.period_min,
                      "period_max": sherlock_target.period_max, "n_transits_min": transits_min_count,
                      "T0_fit_margin": sherlock_target.t0_fit_margin, "show_progress_bar": False,
                      "use_threads": sherlock_target.cpu_cores, "oversampling_factor": oversampling,
                      "duration_grid_step": sherlock_target.duration_grid_step,
                      "period_grid": period_grid}
        if star_info.ld_coefficients is not None:
            power_args["u"] = star_info.ld_coefficients
        power_args["R_star"] = star_info.radius
        power_args["R_star_min"] = star_info.radius_min
        power_args["R_star_max"] = star_info.radius_max
        power_args["M_star"] = star_info.mass
        power_args["M_star_min"] = star_info.mass_min
        power_args["M_star_max"] = star_info.mass_max
        power_args['use_gpu'] = sherlock_target.search_engine == 'gpu' or sherlock_target.search_engine == 'gpu_approximate'
        power_args['gpu_approximate'] = sherlock_target.search_engine == 'gpu_approximate'
        if sherlock_target.custom_transit_template is not None:
            power_args["transit_template_generator"] = sherlock_target.custom_transit_template
        results = model.power(**power_args)
        depths = (1 - results.transit_depths) * 1000
        depths_err = results.transit_depths_uncertainties * 1000
        if results.T0 != 0:
            depths_calc = results.transit_depths[~np.isnan(results.transit_depths)]
            depth = (1. - np.mean(depths_calc)) * 1000
            depth_err = np.sqrt(np.nansum([depth_err ** 2 for depth_err in depths_err])) / len(depths_err)
        else:
            t0s = results.transit_times
            depth = results.transit_depths
            depth_err = np.nan
        t0s = np.array(results.transit_times)
        in_transit = tls.transit_mask(time, results.period, results.duration, results.T0)
        transit_count = results.distinct_transit_count
        border_score = compute_border_score(time, results, in_transit, cadence)
        # Recalculating duration because of tls issue https://github.com/hippke/tls/issues/83
        intransit_folded_model = np.where(results['model_folded_model'] < 1.)[0]
        if len(intransit_folded_model) > 0:
            duration = results['period'] * (results['model_folded_phase'][intransit_folded_model[-1]]
                                            - results['model_folded_phase'][intransit_folded_model[0]])
        else:
            duration = results['duration']
        harmonic = self.__is_harmonic(results, run_results, report, detrend_source_period)
        harmonic_power = harmonic_spectrum(results['periods'], results.power)
        return TransitResult(power_args, results, results.period, results.period_uncertainty, duration,
                             results.T0, t0s, depths, depths_err, depth, depth_err, results.odd_even_mismatch,
                             (1 - results.depth_mean_even[0]) * 1000, (1 - results.depth_mean_odd[0]) * 1000, transit_count,
                             results.snr, results.SDE, results.FAP, border_score, in_transit, harmonic,
                             harmonic_power)

    def __calculate_planet_radius(self, star_info, depth):
        rp = np.nan
        try:
            rp = star_info.radius * math.sqrt(depth / 1000) / 0.0091577
        except ValueError as e:
            logging.error("Planet radius could not be calculated: depth=%s, star_radius=%s", depth, star_info.radius)
            logging.exception("Planet radius could not be calculated")
            print(e)
        return rp

    def __is_harmonic(self, tls_results, run_results, report, detrend_source_period):
        scales = [0.25, 0.5, 1, 2, 4]
        if detrend_source_period is not None:
            rotator_scale = round(tls_results.period / detrend_source_period, 2)
            rotator_harmonic = np.array(np.argwhere((np.array(scales) > rotator_scale - 0.02) & (np.array(scales) < rotator_scale + 0.02))).flatten()
            if len(rotator_harmonic) > 0:
                return str(scales[rotator_harmonic[0]]) + "*source"
        period_scales = [tls_results.period / round(item["period"], 2) for item in report]
        for key, period_scale in enumerate(period_scales):
            period_harmonic = np.array(np.argwhere((np.array(scales) > period_scale - 0.02) & (np.array(scales) < period_scale + 0.02))).flatten()
            if len(period_harmonic) > 0:
                period_harmonic = scales[period_harmonic[0]]
                return str(period_harmonic) + "*SOI" + str(key + 1)
        period_scales = [round(tls_results.period / run_results[key].period, 2) if run_results[key].period > 0 else 0 for key in run_results]
        for key, period_scale in enumerate(period_scales):
            period_harmonic = np.array(np.argwhere(
                (np.array(scales) > period_scale - 0.02) & (np.array(scales) < period_scale + 0.02))).flatten()
            if len(period_harmonic) > 0 and period_harmonic[0] != 2:
                period_harmonic = scales[period_harmonic[0]]
                return str(period_harmonic) + "*this(" + str(key) + ")"
        return "-"

    def __trim_axs(self, axs, N):
        [axis.remove() for axis in axs.flat[N:]]
        return axs.flat[:N]

    def __apply_mask_from_transit_results(self, sherlock_target, time, lcs, transit_results, best_signal_index):
        intransit = tls.transit_mask(time, transit_results[best_signal_index].period,
                                     2 * transit_results[best_signal_index].duration, transit_results[best_signal_index].t0)
        if sherlock_target.mask_mode == 'subtract':
            model_flux, model_flux_edges, model_flux_binnumber = stats.binned_statistic(
                transit_results[best_signal_index].results.model_lightcurve_time,
                transit_results[best_signal_index].results.model_lightcurve_model, statistic='mean', bins=len(intransit))
            for flux in lcs:
                flux[intransit] = flux[intransit] + np.full(len(flux[intransit]), 1) - model_flux[intransit]
                flux[intransit] = np.full(len(flux[intransit]), 1)
            clean_time = time
            clean_lcs = lcs
        else:
            clean_lcs = []
            for key, flux in enumerate(lcs):
                flux[intransit] = np.nan
                clean_time = time
                clean_lcs.append(flux)
        return clean_time, np.array(clean_lcs)

    def run_multiprocessing(self, n_processors, func, func_input):
        with Pool(processes=n_processors) as pool:
            return pool.map(func, func_input)

    class KoiInput:
        def __init__(self, star_id, kic_id):
            self.star_id = star_id
            self.kic_id = kic_id



import logging
import os
import shutil
import time
import traceback
from argparse import ArgumentParser
from lcbuilder.constants import USER_HOME_ELEANOR_CACHE
from lcbuilder.eleanor_manager import EleanorManager
from sherlockpipe.ois.OisManager import OisManager
from eleanor.maxsector import maxsector


class Updater:
    """
    This class should be used to download SHERLOCK data as objects of interest, ELEANOR or LATTE metadata. It stores
    timestamps to remind when was the last time it updated and if called again, it could do nothing if it reads very
    recent operations.
    """
    def __init__(self, cache_dir=os.path.expanduser('~') + "/"):
        self.cache_dir = cache_dir

    def update(self, clean, ois, force):
        """
        The main method of the update task. Reloads OIs, ELEANOR and LATTE metadata.

        :param clean: Specifies whether the ELEANOR data should be wiped and downloaded again.
        :param ois: Specifies whether the OIs metadata is the only one to be refreshed (ignoring ELEANOR and LATTE)
        :param force: Specifies whether the last download timestamp should be ignored and proceed as if a refresh was needed.
        """
        ois_manager = OisManager(self.cache_dir)
        timestamp_ois_path = os.path.join(self.cache_dir, '.sherlockpipe/timestamp_ois.txt')
        ois_timestamp = 0
        force = force or clean
        if os.path.exists(timestamp_ois_path):
            with open(timestamp_ois_path, 'r+') as f:
                ois_timestamp = f.read()
        if force or time.time() - float(ois_timestamp) > 3600 * 24 * 7:
            logging.info("------------------ Reloading TOIs ------------------")
            ois_manager.update_tic_csvs()
            logging.info("------------------ Reloading KOIs ------------------")
            ois_manager.update_kic_csvs()
            logging.info("------------------ Reloading EPICs ------------------")
            ois_manager.update_epic_csvs()
            with open(os.path.join(os.path.expanduser('~'), '.sherlockpipe/timestamp_ois.txt'), 'w+') as f:
                f.write(str(time.time()))
        print("DONE")


if __name__ == '__main__':
    ap = ArgumentParser(description='Updater of SHERLOCK PIPEline metadata')
    ap.add_argument('--clean', dest='clean', action='store_true', help="Whether to remove all data and download it again.")
    ap.add_argument('--only_ois', dest='ois', action='store_true', help="Whether to only refresh objects of interest.")
    ap.add_argument('--force', dest='force', action='store_true', help="Whether to ignore update timestamps and do everything again.")
    args = ap.parse_args()
    Updater(USER_HOME_ELEANOR_CACHE).update(args.clean, args.ois, args.force)



from argparse import ArgumentParser
from sherlockpipe.system_stability.run import run_stability


def stability_args_parse(args=None):
    ap = ArgumentParser(description='Validation of system stability')
    ap.add_argument('--object_dir', help="If the object directory is not your current one you need to provide the "
                                         "ABSOLUTE path", required=False)
    ap.add_argument('--max_ecc', type=float, default=0.1,
                    help="Upper limit for the eccentricity grid.",
                    required=False)
    ap.add_argument('--properties', help="The YAML file to be used as input.", required=False)
    ap.add_argument('--cpus', type=int, default=4, help="The number of CPU cores to be used.", required=False)
    ap.add_argument('--period_bins', type=int, default=1, help="The number of period bins to use.", required=False)
    ap.add_argument('--ecc_bins', type=int, default=1, help="The number of eccentricity bins to use.", required=False)
    ap.add_argument('--inc_bins', type=int, default=1, help="The number of inclination bins to use.", required=False)
    ap.add_argument('--omega_bins', type=int, default=1, help="The number of argument of periastron bins to use.",
                    required=False)
    ap.add_argument('--mass_bins', type=int, default=1, help="The number of mass bins to use.", required=False)
    ap.add_argument('--star_mass_bins', type=int, default=1, help="The number of star mass bins to use.",
                    required=False)
    ap.add_argument('--years', type=int, default=500, help="The number of years to integrate (for MEGNO).",
                    required=False)
    ap.add_argument('--free_params', type=str, default=None, help="The parameters to be entirely sampled, separated by "
                                                                  "commas. E.g. 'eccentricity,omega'", required=False)
    return ap.parse_args(args)


if __name__ == '__main__':
    args = stability_args_parse()
    run_stability(args)



import logging
import math
import re
import shutil
from pathlib import Path

import alexfitter
import numpy as np
import pandas as pd
import os
from os import path
import matplotlib.pyplot as plt
from sherlockpipe.loading.tool_with_candidate import ToolWithCandidate


resources_dir = path.join(path.dirname(__file__))
resources_dir = str(Path(resources_dir).parent.absolute())


class Fitter(ToolWithCandidate):
    """
    Used to run a bayesian fit on the Sherlock search results.
    """

    def __init__(self, object_dir, fit_dir, only_initial, is_candidate_from_search, candidates_df, mcmc=False,
                 detrend=None):
        super().__init__(is_candidate_from_search, candidates_df)
        self.object_dir = os.getcwd() if object_dir is None else object_dir
        self.data_dir = fit_dir
        self.only_initial = only_initial
        self.mcmc = mcmc
        self.detrend = detrend

    def fit(self, candidate_df, star_df, cpus, allesfit_dir, tolerance, fit_orbit):
        """
        Main method to run the allesfitter fit.

        :param candidate_df: the candidates dataframe.
        :param star_df: the star dataframe.
        :param cpus: the number of cpus to be used.
        :param allesfit_dir: the directory for the fit to be run.
        :param tolerance: the nested sampling tolerance threshold.
        :param fit_orbit: whether to fit eccentricity and arg. of periastron.
        """
        logging.info("Preparing fit files")
        sherlock_star_file = self.object_dir + "/params_star.csv"
        star_file = allesfit_dir + "/params_star.csv"
        params_file = allesfit_dir + "/params.csv"
        settings_file = allesfit_dir + "/settings.csv"
        lc_file = Fitter.select_lc_file(candidate_df)
        run = int(candidate_df['number'].iloc[0])
        lc = pd.read_csv(self.object_dir + lc_file, header=0)
        time, flux, flux_err = lc["#time"].values, lc["flux"].values, lc["flux_err"].values
        time, flux, flux_err = self.mask_previous_candidates(time, flux, flux_err, run)
        lc = pd.DataFrame(columns=['#time', 'flux', 'flux_err'])
        lc['#time'] = time
        lc['flux'] = flux
        lc['flux_err'] = flux_err
        lc.to_csv(allesfit_dir + "/lc.csv", index=False)
        if os.path.exists(sherlock_star_file) and os.path.isfile(sherlock_star_file):
            shutil.copyfile(sherlock_star_file, star_file)
        shutil.copyfile(resources_dir + "/resources/allesfitter/settings.csv", settings_file)
        fit_width = Fitter.select_fit_width(candidate_df)
        # TODO replace sherlock properties from allesfitter files
        # TODO only use params_star when the star mass or radius was not assumed
        logging.info("Preparing fit properties")
        with open(settings_file, 'r+') as f:
            text = f.read()
            text = re.sub('\\${sherlock:cores}', str(cpus), text)
            text = re.sub('\\${sherlock:fit_width}', str(fit_width), text)
            text = re.sub('\\${sherlock:fit_ttvs}', "False", text)
            text = re.sub('\\${sherlock:names}', ' '.join(candidate_df["name"].astype('str')), text)
            text = re.sub('\\${sherlock:tolerance}', str(tolerance), text)
            if self.detrend == 'hybrid_spline':
                detrend_param = "baseline_flux_lc,hybrid_spline"
            elif self.detrend == 'gp':
                detrend_param = 'baseline_flux_lc,sample_GP_Matern32'
            else:
                detrend_param = ''
            text = re.sub('\\${sherlock:detrend}', detrend_param, text)
            f.seek(0)
            f.write(text)
            f.truncate()
        with open(params_file, 'w') as f:
            f.write(Fitter.fill_candidates_params(candidate_df, star_df, fit_orbit, self.detrend))
            f.truncate()
        logging.info("Running initial guess")
        try:
            alexfitter.show_initial_guess(allesfit_dir)
        except Exception as e:
            logging.exception(str(e))
        # TODO fix custom_plot for all candidates
        # Fitter.custom_plot(candidate_row["name"], candidate_row["period"], fit_width, allesfit_dir, "initial_guess")
        if not self.only_initial:
            logging.info("Preparing bayesian fit")
            if not self.mcmc:
                logging.info("Running dynamic nested sampling")
                try:
                    alexfitter.ns_fit(allesfit_dir)
                    alexfitter.ns_output(allesfit_dir)
                except Exception as e:
                    logging.exception(str(e))
            elif self.mcmc:
                logging.info("Running MCMC")
                try:
                    alexfitter.mcmc_fit(allesfit_dir)
                    alexfitter.mcmc_output(allesfit_dir)
                except Exception as e:
                    logging.exception(str(e))
            logging.info("Generating custom plots")
            # TODO fix custom_plot for all candidates
            # Fitter.custom_plot(candidate_row["name"], candidate_row["period"], fit_width, allesfit_dir, "posterior")

    @staticmethod
    def select_lc_file(candidate_df):
        """
        Chooses the lightcurve file to be used for the fit. If the number of candidates is greater than 1, it will use
        the PDCSAP_flux by default. If only one candidate is used, the candidate is fit against its better Sherlock
        detrend.

        :param candidate_df: the candidate dataframe
        :return: the lightcurve file path to be used.
        """
        if len(candidate_df) > 1:
            lc_file = "/lc.csv"
        else:
            candidate_row = candidate_df.iloc[0]
            lc_file = "/lc_" + str(candidate_row["curve"]) + ".csv"
        return lc_file

    @staticmethod
    def select_fit_width(candidate_df):
        """
        Calculates the window to be used for the fit around the transits.

        :param candidate_df: the candidates dataframe
        :return: the maximum calculated fit_width
        """
        fit_width = 0
        for key, row in candidate_df.iterrows():
            if row["duration"] is not None:
                row_fit_width = float(row["duration"]) / 60 / 24 * 7
                fit_width = fit_width if fit_width > row_fit_width else row_fit_width
        return fit_width

    @staticmethod
    def custom_plot(name, period, fit_width, allesfit_dir, mode="posterior"):
        """
        Creates a custom fit plot from the allesfitter data

        :param name: the candidate name
        :param period: the final period
        :param fit_width: the fit_width window
        :param allesfit_dir: the directory where allesfitter data is stred
        :param mode: the allesfitter plot model
        """
        allesclass = alexfitter.allesclass(allesfit_dir)
        baseline_width = fit_width * 24
        baseline_to_period = fit_width / period
        fig, axes = plt.subplots(2, 3, figsize=(18, 6), gridspec_kw={'height_ratios': [3, 1]}, sharex='col')
        fig.subplots_adjust(left=0.5, right=1.5, hspace=0)
        style = 'full'
        allesclass.plot('lc', name, style, ax=axes[0][0], mode=mode)
        axes[0][0].set_title('lc, ' + style)
        allesclass.plot('lc', name, style + '_residuals', ax=axes[1][0], mode=mode)
        axes[1][0].set_title('')
        style = 'phase'
        allesclass.plot('lc', name, style, ax=axes[0][1], mode=mode, zoomwindow=baseline_to_period)
        axes[0][1].set_title('lc, ' + style)
        axes[0][1].set_xlim([- baseline_to_period / 2, baseline_to_period / 2])
        allesclass.plot('lc', name, style + '_residuals', ax=axes[1][1], mode=mode, zoomwindow=baseline_to_period)
        axes[1][1].set_title('')
        axes[1][1].set_xlim([- baseline_to_period / 2, baseline_to_period / 2])
        style = 'phasezoom'
        allesclass.plot('lc', name, style, ax=axes[0][2], mode=mode, zoomwindow=baseline_width)
        axes[0][2].set_title('lc, ' + style)
        axes[0][2].set_xlim([- baseline_width / 2, baseline_width / 2])
        allesclass.plot('lc', name, style + '_residuals', ax=axes[1][2], mode=mode, zoomwindow=baseline_width)
        axes[1][2].set_title('')
        axes[1][2].set_xlim([- baseline_width / 2, baseline_width / 2])
        fig.savefig(allesfit_dir + '/results/ns_' + mode + '_' + name + '_custom.pdf', bbox_inches='tight')
        style = ['phasezoom_occ']
        style = ['phase_variations']

    @staticmethod
    def fill_candidates_params(candidate_df, star_df, fit_orbit, detrend_mode):
        """
        Fills the candidate planets initial parameters and their distribution to be fit.

        :param candidate_df: the candidates dataframe
        :param star_df: the star dataframe
        :param fit_orbit: whether to fit eccentricity and arg. of periastron.
        :param detrend_mode: type of detrend to be used
        :return: the allesfitter candidates parameters
        """
        candidate_priors_text = ""
        for key, row in candidate_df.iterrows():
            candidate_priors_text = candidate_priors_text + \
                                    Fitter.fill_candidate_params(row, star_df, fit_orbit)
        return candidate_priors_text + Fitter.fill_instrument_params(star_df, detrend_mode)

    @staticmethod
    def fill_candidate_params(candidate_row, star_df, fit_orbit):
        """
        Fills the candidate planet initial parameters and their distribution to be fit.

        :param candidate_row: the candidate row from the dataframe
        :param star_df: the star dataframe
        :param fit_orbit: whether to fit eccentricity and arg. of periastron.
        :return: the allesfitter candidate parameters
        """
        candidate_params = """#name,value,fit,bounds,label,unit
#companion b astrophysical params,,,,,
${sherlock:name}_rr,${sherlock:rp_rs},1,uniform ${sherlock:rp_rs_min} ${sherlock:rp_rs_max},$R_b / R_\star$,
${sherlock:name}_rsuma,${sherlock:sum_rp_rs_a},1,uniform ${sherlock:sum_rp_rs_a_min} ${sherlock:sum_rp_rs_a_max},$(R_\star + R_b) / a_b$,
${sherlock:name}_cosi,0.0,1,uniform 0.0 0.06,$\cos{i_b}$,
${sherlock:name}_epoch,${sherlock:t0},1,uniform ${sherlock:t0_min} ${sherlock:t0_max},$T_{0;b}$,$\mathrm{BJD}$
${sherlock:name}_period,${sherlock:period},1,uniform ${sherlock:period_min} ${sherlock:period_max},$P_b$,$\mathrm{d}$
${sherlock:name}_f_c,0.0,${sherlock:fit_orbit},uniform -1.0 1.0,$\sqrt{e_b} \cos{\omega_b}$,
${sherlock:name}_f_s,0.0,${sherlock:fit_orbit},uniform -1.0 1.0,$\sqrt{e_b} \sin{\omega_b}$,
"""
        candidate_params = re.sub('\\${sherlock:t0}', str(candidate_row["t0"]), candidate_params)
        candidate_params = re.sub('\\${sherlock:t0_min}', str(candidate_row["t0"] - 0.02), candidate_params)
        candidate_params = re.sub('\\${sherlock:t0_max}', str(candidate_row["t0"] + 0.02), candidate_params)
        candidate_params = re.sub('\\${sherlock:period}', str(candidate_row["period"]), candidate_params)
        candidate_params = re.sub('\\${sherlock:period_min}', str(candidate_row["period"] - candidate_row["per_err"]),
                                  candidate_params)
        candidate_params = re.sub('\\${sherlock:period_max}', str(candidate_row["period"] + candidate_row["per_err"]),
                                  candidate_params)
        candidate_params = re.sub('\\${sherlock:fit_orbit}', str(int(fit_orbit)), candidate_params)
        rp_rs = candidate_row["rp_rs"] if candidate_row["rp_rs"] != "-" else 0.1
        depth = candidate_row["depth"] / 1000
        # TODO calculate depth error in SHERLOCK maybe given the std deviation from the depths or even using the
        #  residuals
        depth_err = depth * 0.2
        rp_rs_err = 0.5 / math.sqrt(depth) * depth_err
        candidate_params = re.sub('\\${sherlock:rp_rs}', str(rp_rs), candidate_params)
        rp_rs_min = rp_rs - 2 * rp_rs_err
        rp_rs_min = rp_rs_min if rp_rs_min > 0 else 0.0000001
        rp_rs_max = rp_rs + 2 * rp_rs_err
        rp_rs_max = rp_rs_max if rp_rs_max < 1 else 0.9999
        candidate_params = re.sub('\\${sherlock:rp_rs_min}', str(rp_rs_min), candidate_params)
        candidate_params = re.sub('\\${sherlock:rp_rs_max}', str(rp_rs_max), candidate_params)
        G = 6.674e-11
        mstar_kg = star_df.iloc[0]["M_star"] * 2e30
        mstar_kg_lerr = star_df.iloc[0]["M_star_lerr"] * 2e30
        mstar_kg_uerr = star_df.iloc[0]["M_star_uerr"] * 2e30
        rstar_au = star_df.iloc[0]['R_star'] * 0.00465047
        rstar_lerr_au = star_df.iloc[0]["R_star_lerr"] * 0.00465047
        rstar_uerr_au = star_df.iloc[0]["R_star_uerr"] * 0.00465047
        per = candidate_row["period"] * 24 * 3600
        per_err = candidate_row["per_err"] * 24 * 3600
        a = (G * mstar_kg * per ** 2 / 4. / (np.pi ** 2)) ** (1. / 3.)
        a_au = a / 1.496e11
        sum_rp_rs_a = (np.sqrt(depth) + 1.) * rstar_au / a_au
        sum_rp_rs_a_lerr = np.sqrt(
            np.square(0.5 * depth_err / depth) + np.square(rstar_lerr_au / rstar_au) + np.square(
                2 * per_err / 3. / per) + np.square(mstar_kg_lerr / 3. / mstar_kg)) * sum_rp_rs_a
        sum_rp_rs_a_uerr = np.sqrt(
            np.square(0.5 * depth_err / depth) + np.square(rstar_uerr_au / rstar_au) + np.square(
                2 * per_err / 3. / per) + np.square(mstar_kg_uerr / 3. / mstar_kg)) * sum_rp_rs_a
        sum_rp_rs_a_min = sum_rp_rs_a - 2 * sum_rp_rs_a_lerr
        sum_rp_rs_a_max = sum_rp_rs_a + 2 * sum_rp_rs_a_uerr
        sum_rp_rs_a_min = sum_rp_rs_a_min if sum_rp_rs_a_min > 0 else 0.0000001
        candidate_params = re.sub('\\${sherlock:sum_rp_rs_a}', str(sum_rp_rs_a), candidate_params)
        candidate_params = re.sub('\\${sherlock:sum_rp_rs_a_min}', str(sum_rp_rs_a_min), candidate_params)
        candidate_params = re.sub('\\${sherlock:sum_rp_rs_a_max}', str(sum_rp_rs_a_max), candidate_params)
        candidate_params = re.sub('\\${sherlock:name}', str(candidate_row["name"]), candidate_params)
        return candidate_params

    @staticmethod
    def fill_instrument_params(star_df, detrend_mode):
        """
        Fills the star and systematics information for each instrument (so far only one)

        :param star_df: the star dataframe
        :param detrend_mode: type of detrend to be used
        :return: the allesfitter instrument params
        """
        instrument_params = """#dilution per instrument,,,,,
dil_lc,0.0,0,trunc_normal 0 1 0.0 0.0,$D_\mathrm{0; lc}$,
#limb darkening coefficients per instrument,,,,,
host_ldc_q1_lc,${sherlock:ld_a},1,uniform 0.0 1.0,$q_{1; \mathrm{lc}}$,
host_ldc_q2_lc,${sherlock:ld_b},1,uniform 0.0 1.0,$q_{2; \mathrm{lc}}$,
#surface brightness per instrument and companion,,,,,
host_sbratio_lc,0.0,0,trunc_normal 0 1 0.0 0.0,$J_{b; \mathrm{lc}}$,
#albedo per instrument and companion,,,,,
host_geom_albedo_lc,0.0,0,trunc_normal 0 1 0.0 0.0,$A_{\mathrm{geom}; host; \mathrm{lc}}$,
${sherlock:name}_geom_albedo_lc,0.0,0,trunc_normal 0 1 0.0 0.0,$A_{\mathrm{geom}; b; \mathrm{lc}}$,
#gravity darkening per instrument and companion,,,,,
host_gdc_lc,0.0,0,trunc_normal 0 1 0.0 0.0,$Grav. dark._{b; \mathrm{lc}}$,
#spots per instrument and companion,,,,,
#errors per instrument,
ln_err_flux_lc,-7.0,1,uniform -15.0 0.0,$\log{\sigma_\mathrm{lc}}$,$\log{ \mathrm{rel. flux.} }$
#baseline per instrument,
${sherlock:baseline_params}
"""
        if star_df.iloc[0]["ld_a"] is not None and not np.isnan(star_df.iloc[0]["ld_a"]):
            instrument_params = re.sub('\\${sherlock:ld_a}', str(star_df.iloc[0]["ld_a"]), instrument_params)
            instrument_params = re.sub('\\${sherlock:ld_b}', str(star_df.iloc[0]["ld_b"]), instrument_params)
        else:
            instrument_params = re.sub('\\${sherlock:ld_a}', "0.5", instrument_params)
            instrument_params = re.sub('\\${sherlock:ld_b}', "0.5", instrument_params)
        if detrend_mode == 'gp':
            baseline_params = 'baseline_gp_matern32_lnsigma_flux_lc,0.0,1,uniform -15.0 15.0,' \
                              '$\mathrm{gp: \ln{\sigma} (lc)}$,\n' + \
                              'baseline_gp_matern32_lnrho_flux_lc,0.0,1,uniform -15.0 15.0,' \
                              '$\mathrm{gp: \ln{\rho} (lc)}$,'
        else:
            baseline_params = ""
        instrument_params = re.sub('\\${sherlock:baseline_params}', baseline_params, instrument_params)
        return instrument_params


import unittest

import pkg_resources

from sherlockpipe.loading import common


class TestCommon(unittest.TestCase):
    MODULE_PATH = pkg_resources.resource_filename("sherlockpipe.tests.resources", "RandomSignalSelector.py")
    YAML_PATH = pkg_resources.resource_filename("sherlockpipe.tests.resources", "user-properties.yaml")

    def test_load_module(self):
        loaded_module = common.load_module(self.MODULE_PATH)
        self.assertIsNotNone(loaded_module)

    def test_extract_custom_class(self):
        loaded_class = common.extract_custom_class(self.MODULE_PATH)
        self.assertIsNotNone(loaded_class)

    def test_get_from_dict(self):
        properties = common.load_from_yaml(self.YAML_PATH)
        targets = common.get_from_dict(properties, 'TARGETS')
        snr_min = common.get_from_dict(properties, 'SNR_MIN')
        self.assertIn('TIC 305048087', targets)
        self.assertIsInstance(targets, dict)
        self.assertIsInstance(snr_min, (int, float))

    def test_get_from_dict_or_default(self):
        properties = common.load_from_yaml(self.YAML_PATH)
        expected_value = 'mozart'
        value = common.get_from_dict_or_default(properties, 'AUTHOR', expected_value)
        self.assertEqual(expected_value, value)

    def test_get_from_user_or_config(self):
        properties = common.load_from_yaml(self.YAML_PATH)
        target = common.get_from_dict(properties, 'TARGETS')
        target = common.get_from_dict(target, 'TIC 305048087')
        expected_root_value = 5
        expected_value = 6
        value = common.get_from_dict(properties, 'SNR_MIN')
        self.assertEqual(expected_root_value, value)
        value = common.get_from_user_or_config(target, properties, 'SNR_MIN')
        self.assertEqual(expected_value, value)

    def test_get_from_user_or_config_or_default(self):
        properties = common.load_from_yaml(self.YAML_PATH)
        target = common.get_from_dict(properties, 'TARGETS')
        target = common.get_from_dict(target, 'TIC 305048087')
        expected_root_value = None
        expected_inner_value = None
        expected_value = 'ION'
        value = common.get_from_dict(properties, 'INVENT')
        self.assertEqual(expected_root_value, value)
        value = common.get_from_user_or_config(target, properties, 'INVENT')
        self.assertEqual(expected_inner_value, value)
        value = common.get_from_user_or_config_or_default(target, properties, 'INVENT', expected_value)
        self.assertEqual(expected_value, value)

    def test_load_from_yaml(self):
        properties = common.load_from_yaml(self.YAML_PATH)
        self.assertIsInstance(properties, dict)
        self.assertIn('TARGETS', properties)



import numpy as np
from scipy.ndimage.interpolation import shift


def compute_border_score(time, result, intransit, cadence):
    shift_cadences = 3600 / cadence
    edge_limit_days = 0.05
    transit_depths = np.nan_to_num(result.transit_depths)
    transit_depths = np.zeros(1) if type(transit_depths) is not np.ndarray else transit_depths
    transit_depths = transit_depths[transit_depths > 0] if len(transit_depths) > 0 else []
    # a=a[np.where([i for i, j in groupby(intransit)])]
    border_score = 0
    if len(transit_depths) > 0:
        shifted_transit_points = shift(intransit, shift_cadences, cval=np.nan)
        inverse_shifted_transit_points = shift(intransit, -shift_cadences, cval=np.nan)
        intransit_shifted = intransit | shifted_transit_points | inverse_shifted_transit_points
        time_edge_indexes = np.where(abs(time[:-1] - time[1:]) > edge_limit_days)[0]
        time_edge = np.full(len(time), False)
        time_edge[time_edge_indexes] = True
        time_edge[0] = True
        time_edge[len(time_edge) - 1] = True
        transits_in_edge = intransit_shifted & time_edge
        transits_in_edge_count = len(transits_in_edge[transits_in_edge])
        border_score = 1 - transits_in_edge_count / len(transit_depths)
    return border_score if border_score >= 0 else 0

def find_nearest(array, value):
    idx = (np.abs(array - value)).argmin()
    return idx, array[idx]

def harmonic_spectrum(periods, spectrum):
    harmonics = [1 / 4, 1 / 3, 1 / 2, 1, 2, 3, 4]
    harmonic_spectrum = np.zeros(len(periods))
    for index, period in enumerate(periods):
        harmonic_sde = 0
        non_nan_harmonics_count = 0
        for harmonic in harmonics:
            expected_harmonic_period = period * harmonic
            harmonic_index, harmonic_period = find_nearest(periods, expected_harmonic_period)
            harmonic_index = harmonic_index if np.abs(1 - harmonic_period / expected_harmonic_period) < 0.01 else np.nan
            if not np.isnan(harmonic_index):
                harmonic_sde = harmonic_sde + spectrum[harmonic_index]
                non_nan_harmonics_count = non_nan_harmonics_count + 1
        harmonic_spectrum[index] = harmonic_sde / non_nan_harmonics_count if non_nan_harmonics_count > 0 else 0
    return harmonic_spectrum



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = "0.42.3"
setuptools.setup(
    name="sherlockpipe", # Replace with your own username
    version=version,
    author="M. Dévora-Pajares & F.J. Pozuelos",
    author_email="mdevorapajares@protonmail.com",
    description="Search for Hints of Exoplanets fRom Lightcurves Of spaCe based seeKers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/franpoz/SHERLOCK",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
    install_requires=['arviz==0.12.1', # Validation required (pytransit, from triceratops)
                        'astroplan==0.8',
                        "astroquery==0.4.6",
                        "alexfitter==1.2.12", # Fit
                        'argparse==1.4.0', # All modules
                        'beautifulsoup4==4.9.3', # Parsing HTML and XML, for OIs extraction
                        "celerite==0.4.0", # Allesfitter dependency
                        "corner==2.2.2", # Allesfitter dependency
                        "dearwatson==0.9.3", # Vetting
                        "dynesty==1.0.1", # Allesfitter dependency
                        "emcee==3.0.2", # Allesfitter dependency
                        "h5py==3.10.0", # Allesfitter dependency
                        "mock==4.0.3",
                        'pdf2image==1.16.2',
                        'pytransit==2.5.21', #Validation
                        'pytz', # Observation plan: Not using version because it gets the DB updated with each release
                        "requests==2.25.1", # OIs management
                        "rebound==3.17.3", # Stability
                        "reproject==0.13.0",
                        "seaborn==0.11.1",
                        'setuptools>=41.0.0',
                        "sklearn==0.0",
                        'statsmodels==0.13.5', # Allesfitter dependency, might conflict with lcbuilder dependency for autocorrelation
                        'timezonefinder==5.2.0', # Observation plan
                        'tqdm==4.56.0',
                        'triceratops==1.0.17', # Validation
                        'uncertainties==3.1.5' # Observation plan
    ]
)



"""Constants used by SHERLOCK"""


import os


USER_HOME_ELEANOR_CACHE = os.path.join(os.path.expanduser('~'), '.eleanor/')
"""The directory where SHERLOCK will store the eleanor internal data"""
MOMENTUM_DUMP_QUALITY_FLAG = 2 ** 5
"""Quality flag value for momentum dumps"""



from argparse import ArgumentParser
from sherlockpipe.observation_plan.run import run_plan

if __name__ == '__main__':
    ap = ArgumentParser(description='Planning of observations from Sherlock objects of interest')
    ap.add_argument('--object_dir',
                    help="If the object directory is not your current one you need to provide the ABSOLUTE path",
                    required=False)
    ap.add_argument("--observatories", help="Csv file containing the observatories coordinates", required=False)
    ap.add_argument('--lat', help="Observer latitude", required=False)
    ap.add_argument('--lon', help="Observer longitude", required=False)
    ap.add_argument('--alt', help="Observer altitude", required=False)
    ap.add_argument('--tz', help="Time zone (in hours offset from UTC)", required=False)
    ap.add_argument('--min_altitude', help="Minimum altitude of the target to be observable", default=23,
                    required=False)
    ap.add_argument('--moon_min_dist', help="Minimum required moon distance for moon minimum illumination.", default=30,
                    required=False)
    ap.add_argument('--moon_max_dist', help="Minimum required moon distance for moon maximum illumination.", default=55,
                    required=False)
    ap.add_argument('--max_days', help="Maximum number of days for the plan to take.", type=int, default=365, required=False)
    ap.add_argument('--transit_fraction', help="Minimum transit fraction to be observable.", type=float, default=0.5,
                    required=False)
    ap.add_argument('--baseline', help="Required baseline (in hours) for the observation.", type=float, default=0.5,
                    required=False)
    ap.add_argument('--since', help="yyyy-mm-dd date since when you want to start the plan (defaults to today).",
                    type=str, default=None,
                    required=False)
    ap.add_argument('--error_sigma', help="Sigma to be used for epoch and period errors.", type=int, default=2,
                    required=False)
    ap.add_argument('--no_error_alert', help="Will not block imprecise observations to be plotted.",
                    action='store_true', required=False)
    ap.add_argument('--time_unit', help="Time unit to be used for the light curve measurements.", default=None,
                    required=False)
    args = ap.parse_args()
    run_plan(args)
